#!/usr/bin/env jupyter
"""
Functions to efficiently merge rows in DataFrames.
"""
import typing as t
from copy import copy

import numpy as np
import pandas as pd
from utils_find_1st import cmp_larger, find_1st

from agora.utils.indexing import compare_indices, validate_association


def apply_merges(data: pd.DataFrame, merges: np.ndarray):
    """Split data in two, one subset for rows relevant for merging and one
    without them. It uses an array of source tracklets and target tracklets
    to efficiently merge them.

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame.
    merges : np.ndarray
        3-D ndarray where dimensions are (X,2,2): nmerges, source-target
        pair and single-cell identifiers, respectively.

    Examples
    --------
    FIXME: Add docs.

    """

    indices = data.index
    if "mother_label" in indices.names:
        indices = indices.droplevel("mother_label")
    valid_merges, indices = validate_association(
        merges, np.array(list(indices))
    )

    # Assign non-merged
    merged = data.loc[~indices]

    # Implement the merges and drop source rows.
    # TODO Use matrices to perform merges in batch
    # for ecficiency
    if valid_merges.any():
        to_merge = data.loc[indices]
        targets, sources = zip(*merges[valid_merges])
        for source, target in zip(sources, targets):
            target = tuple(target)
            to_merge.loc[target] = join_tracks_pair(
                to_merge.loc[target].values,
                to_merge.loc[tuple(source)].values,
            )
        to_merge.drop(map(tuple, sources), inplace=True)

        merged = pd.concat((merged, to_merge), names=data.index.names)
    return merged


def join_tracks_pair(target: np.ndarray, source: np.ndarray) -> np.ndarray:
    """
    Join two tracks and return the new value of the target.
    """
    target_copy = target
    end = find_1st(target_copy[::-1], 0, cmp_larger)
    target_copy[-end:] = source[-end:]
    return target_copy


def group_merges(merges: np.ndarray) -> t.List[t.Tuple]:
    # Return a list where the cell is present as source and target
    # (multimerges)

    sources_targets = compare_indices(merges[:, 0, :], merges[:, 1, :])
    is_multimerge = sources_targets.any(axis=0) | sources_targets.any(axis=1)
    is_monomerge = ~is_multimerge

    multimerge_subsets = union_find(zip(*np.where(sources_targets)))
    merge_groups = [merges[np.array(tuple(x))] for x in multimerge_subsets]

    sorted_merges = list(map(sort_association, merge_groups))

    # Ensure that source and target are at the edges
    return [
        *sorted_merges,
        *[[event] for event in merges[is_monomerge]],
    ]


def union_find(lsts):
    sets = [set(lst) for lst in lsts if lst]
    merged = True
    while merged:
        merged = False
        results = []
        while sets:
            common, rest = sets[0], sets[1:]
            sets = []
            for x in rest:
                if x.isdisjoint(common):
                    sets.append(x)
                else:
                    merged = True
                    common |= x
            results.append(common)
        sets = results
    return sets


def sort_association(array: np.ndarray):
    # Sort the internal associations

    order = np.where(
        (array[:, 0, ..., None] == array[:, 1].T[None, ...]).all(axis=1)
    )

    res = []
    [res.append(x) for x in np.flip(order).flatten() if x not in res]
    sorted_array = array[np.array(res)]
    return sorted_array


def merge_association(
    association: np.ndarray, merges: np.ndarray
) -> np.ndarray:
    grouped_merges = group_merges(merges)

    flat_indices = association.reshape(-1, 2)
    comparison_mat = compare_indices(merges[:, 0], flat_indices)

    valid_indices = comparison_mat.any(axis=0)

    if valid_indices.any():  # Where valid, perform transformation
        replacement_d = {}
        for dataset in grouped_merges:
            for k in dataset:
                replacement_d[tuple(k[0])] = dataset[-1][1]

        flat_indices[valid_indices] = [
            replacement_d[tuple(i)] for i in flat_indices[valid_indices]
        ]

    merged_indices = flat_indices.reshape(-1, 2, 2)
    return merged_indices



import logging
import typing as t
from abc import ABC, abstractmethod
from collections.abc import Iterable
from copy import copy
from pathlib import Path
from time import perf_counter
from typing import Union

from flatten_dict import flatten
from yaml import dump, safe_load

from agora.logging import timer

atomic = t.Union[int, float, str, bool]


class ParametersABC(ABC):
    """
    Defines parameters as attributes and allows parameters to
    be converted to either a dictionary or to yaml.

    No attribute should be called "parameters"!
    """

    def __init__(self, **kwargs):
        """
        Defines parameters as attributes
        """
        assert (
            "parameters" not in kwargs
        ), "No attribute should be named parameters"
        for k, v in kwargs.items():
            setattr(self, k, v)

    def to_dict(self, iterable="null") -> t.Dict:
        """
        Recursive function to return a nested dictionary of the
        attributes of the class instance.
        """
        if isinstance(iterable, dict):
            if any(
                [
                    True
                    for x in iterable.values()
                    if isinstance(x, Iterable) or hasattr(x, "to_dict")
                ]
            ):
                return {
                    k: v.to_dict()
                    if hasattr(v, "to_dict")
                    else self.to_dict(v)
                    for k, v in iterable.items()
                }
            else:
                return iterable
        elif iterable == "null":
            # use instance's built-in __dict__ dictionary of attributes
            return self.to_dict(self.__dict__)
        else:
            return iterable

    def to_yaml(self, path: Union[Path, str] = None):
        """
        Returns a yaml stream of the attributes of the class instance.
        If path is provided, the yaml stream is saved there.

        Parameters
        ----------
        path : Union[Path, str]
            Output path.
        """
        if path:
            with open(Path(path), "w") as f:
                dump(self.to_dict(), f)
        return dump(self.to_dict())

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)

    @classmethod
    def from_yaml(cls, source: Union[Path, str]):
        """
        Returns instance from a yaml filename or stdin
        """
        is_buffer = True
        try:
            if Path(source).exists():
                is_buffer = False
        except Exception as _:
            assert isinstance(source, str), "Invalid source type."

        if is_buffer:
            params = safe_load(source)
        else:
            with open(source) as f:
                params = safe_load(f)
        return cls(**params)

    @classmethod
    def default(cls, **kwargs):
        overriden_defaults = copy(cls._defaults)
        for k, v in kwargs.items():
            overriden_defaults[k] = v
        return cls.from_dict(overriden_defaults)

    def update(self, name: str, new_value):
        """
        Update values recursively
        if name is a dictionary, replace data where existing found or add if not.
        It warns against type changes.

        If the existing structure under name is a dictionary,
        it looks for the first occurrence and modifies it accordingly.

        If a leaf node that is to be changed is a collection, it adds the new elements.
        """

        assert name not in (
            "parameters",
            "params",
        ), "Attribute can't be named params or parameters"

        if name in self.__dict__:
            if check_type_recursive(getattr(self, name), new_value):
                print("Warnings:Type changes are risky")

            if isinstance(getattr(self, name), dict):
                flattened = flatten(self.to_dict())
                names_found = [k for k in flattened.keys() if name in k]
                found_idx = [keys.index(name) for keys in names_found]

                assert len(names_found), f"{name} not found as key."

                keys = None
                if len(names_found) > 1:
                    for level in zip(found_idx, names_found):
                        if level == min(found_idx):
                            keys = level
                            print(
                                f"Warning: {name} was found in multiple keys. Selected {keys}"
                            )
                            break

                else:
                    keys = names_found.pop()

                if keys:
                    current_val = flattened.get(keys, None)
                    # if isinstance(current_val, t.Collection):

            elif isinstance(getattr(self, name), t.Collection):
                add_to_collection(getattr(self, name), new_value)

            elif isinstance(getattr(self, name), set):
                pass  # TODO implement

            new_d = getattr(self, name)
            new_d.update(new_value)
            setattr(self, name, new_d)

        else:
            setattr(self, name, new_value)


def add_to_collection(
    collection: t.Collection, value: t.Union[atomic, t.Collection]
):
    # Adds element(s) in place.
    if not isinstance(value, t.Collection):
        value = [value]
    if isinstance(collection, list):
        collection += value
    elif isinstance(collection, set):
        collection.update(value)


class ProcessABC(ABC):
    """
    Base class for processes.
    Defines parameters as attributes and requires run method to be defined.
    """

    def __init__(self, parameters):
        """
        Arguments
        ---------
        parameters: instance of ParametersABC
        """
        self._parameters = parameters
        # convert parameters to dictionary
        # and then define each parameter as an attribute
        for k, v in parameters.to_dict().items():
            setattr(self, k, v)

    @property
    def parameters(self):
        return self._parameters

    @abstractmethod
    def run(self):
        pass

    def _log(self, message: str, level: str = "warning"):
        # Log messages in the corresponding level
        logger = logging.getLogger("aliby")
        getattr(logger, level)(f"{self.__class__.__name__}: {message}")


def check_type_recursive(val1, val2):
    same_types = True
    if not isinstance(val1, type(val2)) and not all(
        type(x) in (Path, str) for x in (val1, val2)  # Ignore str->path
    ):
        return False
    if not isinstance(val1, t.Iterable) and not isinstance(val2, t.Iterable):
        return isinstance(val1, type(val2))
    elif isinstance(val1, (tuple, list)) and isinstance(val2, (tuple, list)):
        return bool(
            sum([check_type_recursive(v1, v2) for v1, v2 in zip(val1, val2)])
        )
    elif isinstance(val1, dict) and isinstance(val2, dict):
        if not len(val1) or not len(val2):
            return False
        for k in val2.keys():
            same_types = same_types and check_type_recursive(val1[k], val2[k])
    return same_types


class StepABC(ProcessABC):
    """
    Base class that expands on ProcessABC to include tools used by Aliby steps.
    It adds a setup step, logging and benchmarking for time benchmarks.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def _run_tp(self):
        pass

    @timer
    def run_tp(self, tp: int, **kwargs):
        """
        Time and log the timing of a step.
        """
        return self._run_tp(tp, **kwargs)

    def run(self):
        # Replace run with run_tp
        raise Warning("Steps use run_tp instead of run")



"""
Tiler: Divides images into smaller tiles.

The tasks of the Tiler are selecting regions of interest, or tiles, of images - with one trap per tile, correcting for the drift of the microscope stage over time, and handling errors and bridging between the image data and Aliby’s image-processing steps.

Tiler subclasses deal with either network connections or local files.

To find tiles, we use a two-step process: we analyse the bright-field image to produce the template of a trap, and we fit this template to the image to find the tiles' centres.

We use texture-based segmentation (entropy) to split the image into foreground -- cells and traps -- and background, which we then identify with an Otsu filter. Two methods are used to produce a template trap from these regions: pick the trap with the smallest minor axis length and average over all validated traps.

A peak-identifying algorithm recovers the x and y-axis location of traps in the original image, and we choose the approach to template that identifies the most tiles.

The experiment is stored as an array with a standard indexing order of (Time, Channels, Z-stack, X, Y).
"""
import logging
import re
import typing as t
import warnings
from functools import lru_cache
from pathlib import Path

import dask.array as da
import h5py
import numpy as np
from skimage.registration import phase_cross_correlation

from agora.abc import ParametersABC, StepABC
from agora.io.writer import BridgeH5
from aliby.io.image import ImageDummy
from aliby.tile.traps import segment_traps


class Tile:
    """
    Store a tile's location and size.

    Checks to see if the tile should be padded.
    Can export the tile either in OMERO or numpy formats.
    """

    def __init__(self, centre, parent, size, max_size):
        self.centre = centre
        self.parent = parent  # used to access drifts
        self.size = size
        self.half_size = size // 2
        self.max_size = max_size

    def at_time(self, tp: int) -> t.List[int]:
        """
        Return tile's centre by applying drifts.

        Parameters
        ----------
        tp: integer
            Index for the time point of interest.
        """
        drifts = self.parent.drifts
        tile_centre = self.centre - np.sum(drifts[: tp + 1], axis=0)
        return list(tile_centre.astype(int))

    def as_tile(self, tp: int):
        """
        Return tile in the OMERO tile format of x, y, w, h.

        Here x, y are at the bottom left corner of the tile
        and w and h are the tile width and height.

        Parameters
        ----------
        tp: integer
            Index for the time point of interest.

        Returns
        -------
        x: int
            x-coordinate of bottom left corner of tile
        y: int
            y-coordinate of bottom left corner of tile
        w: int
            Width of tile
        h: int
            Height of tile
        """
        x, y = self.at_time(tp)
        # tile bottom corner
        x = int(x - self.half_size)
        y = int(y - self.half_size)
        return x, y, self.size, self.size

    def as_range(self, tp: int):
        """
        Return tile in a range format: two slice objects that can
        be used in arrays.

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        A slice of x coordinates from left to right
        A slice of y coordinates from top to bottom
        """
        x, y, w, h = self.as_tile(tp)
        return slice(x, x + w), slice(y, y + h)


class TileLocations:
    """Store each tile as an instance of Tile."""

    def __init__(
        self,
        initial_location: np.array,
        tile_size: int = None,
        max_size: int = 1200,
        drifts: np.array = None,
    ):
        if drifts is None:
            drifts = []
        self.tile_size = tile_size
        self.max_size = max_size
        self.initial_location = initial_location
        self.tiles = [
            Tile(centre, self, tile_size or max_size, max_size)
            for centre in initial_location
        ]
        self.drifts = drifts

    def __len__(self):
        return len(self.tiles)

    def __iter__(self):
        yield from self.tiles

    @property
    def shape(self):
        """Return numbers of tiles and drifts."""
        return len(self.tiles), len(self.drifts)

    def to_dict(self, tp: int):
        """
        Export initial locations, tile_size, max_size, and drifts
        as a dictionary.

        Parameters
        ----------
        tp: integer
            An index for a time point
        """
        res = dict()
        if tp == 0:
            res["trap_locations"] = self.initial_location
            res["attrs/tile_size"] = self.tile_size
            res["attrs/max_size"] = self.max_size
        res["drifts"] = np.expand_dims(self.drifts[tp], axis=0)
        return res

    def at_time(self, tp: int) -> np.ndarray:
        """Return an array of tile centres (x- and y-coords)."""
        return np.array([tile.at_time(tp) for tile in self.tiles])

    @classmethod
    def from_tiler_init(
        cls, initial_location, tile_size: int = None, max_size: int = 1200
    ):
        """Instantiate from a Tiler."""
        return cls(initial_location, tile_size, max_size, drifts=[])

    @classmethod
    def read_hdf5(cls, file):
        """Instantiate from a h5 file."""
        with h5py.File(file, "r") as hfile:
            tile_info = hfile["trap_info"]
            initial_locations = tile_info["trap_locations"][()]
            drifts = tile_info["drifts"][()].tolist()
            max_size = tile_info.attrs["max_size"]
            tile_size = tile_info.attrs["tile_size"]
        tile_loc_cls = cls(initial_locations, tile_size, max_size=max_size)
        tile_loc_cls.drifts = drifts
        return tile_loc_cls


class TilerParameters(ParametersABC):
    """
    tile_size: int
    ref_channel: str or int
    ref_z: int
    backup_ref_channel int or None, if int indicates the index for reference channel. Used when image does not include metadata, ref_channel is a string and channel names are included in parsed logfiles.

    """

    _defaults = {
        "tile_size": 117,
        "ref_channel": "Brightfield",
        "ref_z": 0,
        "backup_ref_channel": None,
    }


class Tiler(StepABC):
    """
    Divide images into smaller tiles for faster processing.

    Finds tiles and re-registers images if they drift.
    Fetch images from an OMERO server if necessary.

    Uses an Image instance, which lazily provides the data on pixels,
    and, as an independent argument, metadata.
    """

    def __init__(
        self,
        image: da.core.Array,
        metadata: dict,
        parameters: TilerParameters,
        tile_locs=None,
    ):
        """
        Initialise.

        Parameters
        ----------
        image: an instance of Image
        metadata: dictionary
        parameters: an instance of TilerParameters
        tile_locs: (optional)
        """
        super().__init__(parameters)
        self.image = image
        self._metadata = metadata
        self.channels = metadata.get(
            "channels",
            list(range(metadata.get("size_c", 0))),
        )

        self.ref_channel = self.get_channel_index(parameters.ref_channel)
        if self.ref_channel is None:
            self.ref_channel = self.backup_ref_channel

        self.ref_channel = self.get_channel_index(parameters.ref_channel)
        self.tile_locs = tile_locs
        try:
            self.z_perchannel = {
                ch: zsect
                for ch, zsect in zip(self.channels, metadata["zsections"])
            }
        except Exception as e:
            self._log(f"No z_perchannel data: {e}")
        self.tile_size = self.tile_size or min(self.image.shape[-2:])

    @classmethod
    def dummy(cls, parameters: dict):
        """
        Instantiate dummy Tiler from dummy image.

        If image.dimorder exists dimensions are saved in that order.
        Otherwise default to "tczyx".

        Parameters
        ----------
        parameters: dict
            An instance of TilerParameters converted to a dict.
        """
        imgdmy_obj = ImageDummy(parameters)
        dummy_image = imgdmy_obj.get_data_lazy()
        # default to "tczyx" if image.dimorder is None
        dummy_omero_metadata = {
            f"size_{dim}": dim_size
            for dim, dim_size in zip(
                imgdmy_obj.dimorder or "tczyx", dummy_image.shape
            )
        }
        dummy_omero_metadata.update(
            {
                "channels": [
                    parameters["ref_channel"],
                    *(["nil"] * (dummy_omero_metadata["size_c"] - 1)),
                ],
                "name": "",
            }
        )
        return cls(
            imgdmy_obj.data,
            dummy_omero_metadata,
            TilerParameters.from_dict(parameters),
        )

    @classmethod
    def from_image(cls, image, parameters: TilerParameters):
        """
        Instantiate from an Image instance.

        Parameters
        ----------
        image: an instance of Image
        parameters: an instance of TilerPameters
        """
        return cls(image.data, image.metadata, parameters)

    @classmethod
    def from_h5(
        cls,
        image,
        filepath: t.Union[str, Path],
        parameters: t.Optional[TilerParameters] = None,
    ):
        """
        Instantiate from h5 files.

        Parameters
        ----------
        image: an instance of Image
        filepath: Path instance
            Path to a directory of h5 files
        parameters: an instance of TileParameters (optional)
        """
        tile_locs = TileLocations.read_hdf5(filepath)
        metadata = BridgeH5(filepath).meta_h5
        metadata["channels"] = image.metadata["channels"]
        if parameters is None:
            parameters = TilerParameters.default()
        tiler = cls(
            image.data,
            metadata,
            parameters,
            tile_locs=tile_locs,
        )
        if hasattr(tile_locs, "drifts"):
            tiler.n_processed = len(tile_locs.drifts)
        return tiler

    @lru_cache(maxsize=2)
    def get_tc(self, t: int, c: int) -> np.ndarray:
        """
        Load image using dask.

        Assumes the image is arranged as
            no of time points
            no of channels
            no of z stacks
            no of pixels in y direction
            no of pixels in x direction

        Parameters
        ----------
        t: integer
            An index for a time point
        c: integer
            An index for a channel

        Returns
        -------
        full: an array of images
        """
        full = self.image[t, c]
        if hasattr(full, "compute"):  # If using dask fetch images here
            full = full.compute(scheduler="synchronous")

        return full

    @property
    def shape(self):
        """
        Return properties of the time-lapse as shown by self.image.shape
        """
        return self.image.shape

    @property
    def n_processed(self):
        """Return the number of processed images."""
        if not hasattr(self, "_n_processed"):
            self._n_processed = 0
        return self._n_processed

    @n_processed.setter
    def n_processed(self, value):
        self._n_processed = value

    @property
    def n_tiles(self):
        """Return number of tiles."""
        return len(self.tile_locs)

    def initialise_tiles(self, tile_size: int = None):
        """
        Find initial positions of tiles.

        Remove tiles that are too close to the edge of the image
        so no padding is necessary.

        Parameters
        ----------
        tile_size: integer
            The size of a tile.
        """
        initial_image = self.image[0, self.ref_channel, self.ref_z]
        if tile_size:
            half_tile = tile_size // 2
            # max_size is the minimal number of x or y pixels
            max_size = min(self.image.shape[-2:])
            # first time point, reference channel, reference z-position
            # find the tiles
            tile_locs = segment_traps(initial_image, tile_size)
            # keep only tiles that are not near an edge
            tile_locs = [
                [x, y]
                for x, y in tile_locs
                if half_tile < x < max_size - half_tile
                and half_tile < y < max_size - half_tile
            ]
            # store tiles in an instance of TileLocations
            self.tile_locs = TileLocations.from_tiler_init(
                tile_locs, tile_size
            )
        else:
            yx_shape = self.image.shape[-2:]
            tile_locs = [[x // 2 for x in yx_shape]]
            self.tile_locs = TileLocations.from_tiler_init(
                tile_locs, max_size=min(yx_shape)
            )

    def find_drift(self, tp: int):
        """
        Find any translational drift between two images at consecutive
        time points using cross correlation.

        Arguments
        ---------
        tp: integer
            Index for a time point.
        """
        prev_tp = max(0, tp - 1)
        # cross-correlate
        drift, _, _ = phase_cross_correlation(
            self.image[prev_tp, self.ref_channel, self.ref_z],
            self.image[tp, self.ref_channel, self.ref_z],
        )
        # store drift
        if 0 < tp < len(self.tile_locs.drifts):
            self.tile_locs.drifts[tp] = drift.tolist()
        else:
            self.tile_locs.drifts.append(drift.tolist())

    def get_tp_data(self, tp, c) -> np.ndarray:
        """
        Returns all tiles corrected for drift.

        Parameters
        ----------
        tp: integer
            An index for a time point
        c: integer
            An index for a channel

        Returns
        ----------
        Numpy ndarray of tiles with shape (tile, z, y, x)
        """
        tiles = []
        # get OMERO image
        full = self.get_tc(tp, c)
        for tile in self.tile_locs:
            # pad tile if necessary
            ndtile = self.ifoob_pad(full, tile.as_range(tp))
            tiles.append(ndtile)
        return np.stack(tiles)

    def get_tile_data(self, tile_id: int, tp: int, c: int):
        """
        Return a particular tile corrected for drift and padding.

        Parameters
        ----------
        tile_id: integer
            Number of tile.
        tp: integer
            Index of time points.
        c: integer
            Index of channel.

        Returns
        -------
        ndtile: array
            An array of (x, y) arrays, one for each z stack
        """
        full = self.get_tc(tp, c)
        tile = self.tile_locs.tiles[tile_id]
        ndtile = self.ifoob_pad(full, tile.as_range(tp))
        return ndtile

    def _run_tp(self, tp: int):
        """
        Find tiles if they have not yet been found.

        Determine any translational drift of the current image from the
        previous one.

        Arguments
        ---------
        tp: integer
            The time point to tile.
        """
        # assert tp >= self.n_processed, "Time point already processed"
        # TODO check contiguity?
        if self.n_processed == 0 or not hasattr(self.tile_locs, "drifts"):
            self.initialise_tiles(self.tile_size)
        if hasattr(self.tile_locs, "drifts"):
            drift_len = len(self.tile_locs.drifts)
            if self.n_processed != drift_len:
                warnings.warn("Tiler:n_processed and ndrifts don't match")
                self.n_processed = drift_len
        # determine drift
        self.find_drift(tp)
        # update n_processed
        self.n_processed = tp + 1
        # return result for writer
        return self.tile_locs.to_dict(tp)

    def run(self, time_dim=None):
        """
        Tile all time points in an experiment at once.
        """
        if time_dim is None:
            time_dim = 0
        for frame in range(self.image.shape[time_dim]):
            self.run_tp(frame)
        return None

    def get_traps_timepoint(self, *args, **kwargs):
        self._log(
            "get_traps_timepoint is deprecated; get_tiles_timepoint instead."
        )
        return self.get_tiles_timepoint(*args, **kwargs)

    # The next set of functions are necessary for the extraction object
    def get_tiles_timepoint(
        self, tp: int, tile_shape=None, channels=None, z: int = 0
    ) -> np.ndarray:
        """
        Get a multidimensional array with all tiles for a set of channels
        and z-stacks.

        Used by extractor.

        Parameters
        ---------
        tp: int
            Index of time point
        tile_shape: int or tuple of two ints
            Size of tile in x and y dimensions
        channels: string or list of strings
            Names of channels of interest
        z: int
            Index of z-channel of interest

        Returns
        -------
        res: array
            Data arranged as (tiles, channels, time points, X, Y, Z)
        """
        # FIXME add support for sub-tiling a tile
        # FIXME can we ignore z
        if channels is None:
            channels = [0]
        elif isinstance(channels, str):
            channels = [channels]
        # get the data
        res = []
        for c in channels:
            # only return requested z
            val = self.get_tp_data(tp, c)[:, z]
            # starts with the order: tiles, z, y, x
            # returns the order: tiles, C, T, Z, X, Y
            val = np.expand_dims(val, axis=1)
            res.append(val)
        if tile_shape is not None:
            if isinstance(tile_shape, int):
                tile_shape = (tile_shape, tile_shape)
            assert np.all(
                [
                    (tile_size - ax) > -1
                    for tile_size, ax in zip(tile_shape, res[0].shape[-3:-2])
                ]
            )
        return np.stack(res, axis=1)

    @property
    def ref_channel_index(self):
        """Return index of reference channel."""
        return self.get_channel_index(self.parameters.ref_channel)

    def get_channel_index(self, channel: str or int) -> int or None:
        """
        Find index for channel using regex. Returns the first matched string.
        If self.channels is integers (no image metadata) it returns None.
        If channel is integer

        Parameters
        ----------
        channel: string or int
            The channel or index to be used.
        """

        if all(map(lambda x: isinstance(x, int), self.channels)):
            channel = channel if isinstance(channel, int) else None

        if isinstance(channel, str):
            channel = find_channel_index(self.channels, channel)
        return channel

    @staticmethod
    def ifoob_pad(full, slices):
        """
        Return the slices padded if out of bounds.

        Parameters
        ----------
        full: array
            Slice of OMERO image (zstacks, x, y) - the entire position
            with zstacks as first axis
        slices: tuple of two slices
            Delineates indices for the x- and y- ranges of the tile.

        Returns
        -------
        tile: array
            A tile with all z stacks for the given slices.
            If some padding is needed, the median of the image is used.
            If much padding is needed, a tile of NaN is returned.
        """
        # number of pixels in the y direction
        max_size = full.shape[-1]
        # ignore parts of the tile outside of the image
        y, x = [slice(max(0, s.start), min(max_size, s.stop)) for s in slices]
        # get the tile including all z stacks
        tile = full[:, y, x]
        # find extent of padding needed in x and y
        padding = np.array(
            [(-min(0, s.start), -min(0, max_size - s.stop)) for s in slices]
        )
        if padding.any():
            tile_size = slices[0].stop - slices[0].start
            if (padding > tile_size / 4).any():
                # too much of the tile is outside of the image
                # fill with NaN
                tile = np.full((full.shape[0], tile_size, tile_size), np.nan)
            else:
                # pad tile with median value of the tile
                tile = np.pad(tile, [[0, 0]] + padding.tolist(), "median")
        return tile


# FIXME: Refactor to support both channel or index
# self._log below is not defined
def find_channel_index(image_channels: t.List[str], channel: str):
    """
    Access
    """
    for i, ch in enumerate(image_channels):
        found = re.match(channel, ch, re.IGNORECASE)
        if found:
            if len(found.string) - (found.endpos - found.start()):
                logging.getLogger("aliby").log(
                    logging.WARNING,
                    f"Channel {channel} matched {ch} using regex",
                )
            return i


def find_channel_name(image_channels: t.List[str], channel: str):
    """
    Find the name of the channel using regex.

    Parameters
    ----------
    image_channels: list of str
        Channels.
    channel: str
        A regular expression.
    """
    index = find_channel_index(image_channels, channel)
    if index is not None:
        return image_channels[index]



#!/usr/bin/env python3
from pathlib import Path

import h5py
import numpy as np

from agora.io.bridge import groupsort
from agora.io.writer import load_attributes


class DynamicReader:
    group = ""

    def __init__(self, file: str):
        self.file = file
        self.metadata = load_attributes(file)


class StateReader(DynamicReader):
    """
    Analogous to StateWriter:


    Possible cases (and data shapes):
    - max_lbl  (ntraps, 1) -> One int per trap.
    - tp_back, trap, cell_label -> One int per cell_label-timepoint
    - prev_feats -> A fixed number of floats per cell_label-timepoint (default is 9)
    - lifetime, p_was_bud, p_is_mother -> (nTotalCells, 2)  A (Ncells, 2) matrix where the first column is the trap,
        and its index for such trap (+1) is its cell label.
    - ba_cum ->.  (2^n, 2^n, None) 3d array where the lineage score is contained for all traps - traps in the 3rd dimension 3d array where the lineage score is contained for all traps - traps in the 3rd dimension.
        2^n >= ncells, it is kept in powers of two for efficiency.

    """

    data_types = {}
    datatypes = {
        "max_lbl": ((None, 1), np.uint16),
        "tp_back": ((None, 1), np.uint16),
        "trap": ((None, 1), np.int16),
        "cell_lbls": ((None, 1), np.uint16),
        "prev_feats": ((None, None), np.float64),
        "lifetime": ((None, 2), np.uint16),
        "p_was_bud": ((None, 2), np.float64),
        "p_is_mother": ((None, 2), np.float64),
        "ba_cum": ((None, None), np.float64),
    }
    group = "last_state"

    def __init__(self, file: str):
        super().__init__(file)

    def format_tps(self):
        pass

    def format_traps(self):
        pass

    def format_bacum(self):
        pass

    def read_raw(self, key, dtype):
        with h5py.File(self.file, "r") as f:
            raw = f[self.group + "/" + key][()].astype(dtype)

        return raw

    def read_all(self):

        self.raw_data = {
            key: self.read_raw(key, dtype)
            for key, (_, dtype) in self.datatypes.items()
        }

        return self.raw_data

    def reconstruct_states(self, data: dict):
        ntps_back = max(data["tp_back"]) + 1

        from copy import copy

        tpback_as_idx = copy(data["tp_back"])
        trap_as_idx = copy(data["trap"])

        states = {k: {"max_lbl": v} for k, v in enumerate(data["max_lbl"])}
        for val_name in ("cell_lbls", "prev_feats"):
            for k in states.keys():
                if val_name == "cell_lbls":
                    states[k][val_name] = [[] for _ in range(ntps_back)]
                else:
                    states[k][val_name] = [
                        np.zeros(
                            (0, data[val_name].shape[1]), dtype=np.float64
                        )
                        for _ in range(ntps_back)
                    ]

            data[val_name] = list(
                zip(trap_as_idx, tpback_as_idx, data[val_name])
            )
            for k, v in groupsort(data[val_name]).items():
                states[k][val_name] = [
                    np.array([w[0] for w in val])
                    for val in groupsort(v).values()
                ]

        for val_name in ("lifetime", "p_was_bud", "p_is_mother"):
            for k in states.keys():
                states[k][val_name] = np.array([])
            # This contains no time points back
            for k, v in groupsort(data[val_name]).items():
                states[k][val_name] = np.array([val[0] for val in v])

        for trap_id, ba_matrix in enumerate(data["ba_cum"]):
            states[trap_id]["ba_cum"] = np.array(ba_matrix, dtype=np.float64)

        return [val for val in states.values()]

    def get_formatted_states(self):
        return self.reconstruct_states(self.read_all())



import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class TemplateParameters(ParametersABC):
    """
    Parameters
    """

    _defaults = {}


class Template(PostProcessABC):
    """
    Template for process class.
    """

    def __init__(self, parameters: TemplateParameters):
        super().__init__(parameters)

    def run(self):
        pass



#!/usr/bin/env python3
"""
Dataset is a group of classes to manage multiple types of experiments:
 - Remote experiments on an OMERO server (located in src/aliby/io/omero.py)
 - Local experiments in a multidimensional OME-TIFF image containing the metadata
 - Local experiments in a directory containing multiple positions in independent images with or without metadata
"""
import os
import shutil
import time
import typing as t
from abc import ABC, abstractproperty, abstractmethod
from pathlib import Path

from agora.io.bridge import BridgeH5
from aliby.io.image import ImageLocalOME


def dispatch_dataset(expt_id: int or str, **kwargs):
    """
    Find paths to the data.

    Connects to OMERO if data is remotely available.

    Parameters
    ----------
    expt_id: int or str
        To identify the data, either an OMERO ID or an OME-TIFF file or a local directory.

    Returns
    -------
    A callable Dataset instance, either network-dependent or local.
    """
    if isinstance(expt_id, int):
        # data available online
        from aliby.io.omero import Dataset

        return Dataset(expt_id, **kwargs)
    elif isinstance(expt_id, str):
        # data available locally
        expt_path = Path(expt_id)
        if expt_path.is_dir():
            # data in multiple folders
            return DatasetLocalDir(expt_path)
        else:
            # data in one folder as OME-TIFF files
            return DatasetLocalOME(expt_path)
    else:
        raise Warning(f"{expt_id} is an invalid expt_id")


class DatasetLocalABC(ABC):
    """
    Abstract Base class to find local files, either OME-XML or raw images.
    """

    _valid_suffixes = ("tiff", "png", "zarr")
    _valid_meta_suffixes = ("txt", "log")

    def __init__(self, dpath: t.Union[str, Path], *args, **kwargs):
        self.path = Path(dpath)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    @property
    def dataset(self):
        return self.path

    @property
    def name(self):
        return self.path.name

    @property
    def unique_name(self):
        return self.path.name

    @property
    def files(self):
        """Return a dictionary with any available metadata files."""
        if not hasattr(self, "_files"):
            self._files = {
                f: f
                for f in self.path.rglob("*")
                if any(
                    str(f).endswith(suffix)
                    for suffix in self._valid_meta_suffixes
                )
            }
        return self._files

    def cache_logs(self, root_dir):
        """Copy metadata files to results folder."""
        for name, annotation in self.files.items():
            shutil.copy(annotation, root_dir / name.name)
        return True

    @abstractproperty
    def date(self):
        pass

    @abstractmethod
    def get_images(self):
        pass


class DatasetLocalDir(DatasetLocalABC):
    """Find paths to a data set, comprising multiple images in different folders."""

    def __init__(self, dpath: t.Union[str, Path], *args, **kwargs):
        super().__init__(dpath)

    @property
    def date(self):
        """Find date when a folder was created."""
        return time.strftime(
            "%Y%m%d", time.strptime(time.ctime(os.path.getmtime(self.path)))
        )

    def get_images(self):
        """Return a dictionary of folder or file names and their paths.

        FUTURE 3.12 use pathlib is_junction to pick Dir or File
        """
        images = {
            item.name: item
            for item in self.path.glob("*/")
            if item.is_dir()
            and any(
                path
                for suffix in self._valid_suffixes
                for path in item.glob(f"*.{suffix}")
            )
            or item.suffix[1:] in self._valid_suffixes
        }

        return images


class DatasetLocalOME(DatasetLocalABC):
    """Find names of images in a folder, assuming images in OME-TIFF format."""

    def __init__(self, dpath: t.Union[str, Path], *args, **kwargs):
        super().__init__(dpath)
        assert len(
            self.get_images()
        ), f"No valid files found. Formats are {self._valid_suffixes}"

    @property
    def date(self):
        """Get the date from the metadata of the first position."""
        return ImageLocalOME(list(self.get_images().values())[0]).date

    def get_images(self):
        """Return a dictionary with the names of the image files."""
        return {
            f.name: str(f)
            for suffix in self._valid_suffixes
            for f in self.path.glob(f"*.{suffix}")
        }



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'agora': 'src/agora',
 'agora.io': 'src/agora/io',
 'agora.utils': 'src/agora/utils',
 'extraction': 'src/extraction',
 'extraction.core': 'src/extraction/core',
 'extraction.core.functions': 'src/extraction/core/functions',
 'extraction.core.functions.custom': 'src/extraction/core/functions/custom',
 'logfile_parser': 'src/logfile_parser',
 'postprocessor': 'src/postprocessor',
 'postprocessor.core': 'src/postprocessor/core',
 'postprocessor.core.functions': 'src/postprocessor/core/functions',
 'postprocessor.core.multisignal': 'src/postprocessor/core/multisignal',
 'postprocessor.core.processes': 'src/postprocessor/core/processes',
 'postprocessor.core.reshapers': 'src/postprocessor/core/reshapers',
 'postprocessor.routines': 'src/postprocessor/routines'}

packages = \
['agora',
 'agora.io',
 'agora.utils',
 'aliby',
 'aliby.bin',
 'aliby.io',
 'aliby.lineage',
 'aliby.tile',
 'aliby.track',
 'aliby.utils',
 'extraction',
 'extraction.core',
 'extraction.core.functions',
 'extraction.core.functions.custom',
 'logfile_parser',
 'postprocessor',
 'postprocessor.core',
 'postprocessor.core.functions',
 'postprocessor.core.multisignal',
 'postprocessor.core.processes',
 'postprocessor.core.reshapers',
 'postprocessor.routines']

package_data = \
{'': ['*'],
 'aliby.bin': ['2023_03_20_annotation/*'],
 'logfile_parser': ['grammars/*']}

install_requires = \
['Bottleneck>=1.3.5,<2.0.0',
 'GitPython>=3.1.27,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'dask>=2021.12.0,<2022.0.0',
 'flatten-dict>=0.4.2,<0.5.0',
 'gaussianprocessderivatives>=0.1.5,<0.2.0',
 'h5py==2.10',
 'imageio==2.8.0',
 'numpy>=1.21.6',
 'opencv-python>=4.7.0.72,<5.0.0.0',
 'p-tqdm>=1.3.3,<2.0.0',
 'pandas>=1.3.3',
 'pathos>=0.2.8,<0.3.0',
 'py-find-1st>=1.1.5,<2.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'scikit-image>=0.18.1',
 'scikit-learn>=1.0.2',
 'scipy>=1.7.3',
 'tqdm>=4.62.3,<5.0.0',
 'xmltodict>=0.13.0,<0.14.0',
 'zarr>=2.14.0,<3.0.0']

extras_require = \
{'baby': ['aliby-baby>=0.1.17,<0.2.0'], 'omero': ['omero-py>=5.6.2']}

entry_points = \
{'console_scripts': ['aliby-annotate = aliby.bin.annotate:annotate',
                     'aliby-run = aliby.bin.run:run',
                     'aliby-visualise = aliby.bin.visualise:napari_overlay']}

setup_kwargs = {
    'name': 'aliby',
    'version': '0.1.64',
    'description': 'Process and analyse live-cell imaging data',
    'long_description': '# ALIBY (Analyser of Live-cell Imaging for Budding Yeast)\n\n[![docs](https://readthedocs.org/projects/aliby/badge/?version=master)](https://aliby.readthedocs.io/en/latest)\n[![PyPI version](https://badge.fury.io/py/aliby.svg)](https://badge.fury.io/py/aliby)\n[![pipeline](https://gitlab.com/aliby/aliby/badges/master/pipeline.svg?key_text=master)](https://gitlab.com/aliby/aliby/-/pipelines)\n[![dev pipeline](https://gitlab.com/aliby/aliby/badges/dev/pipeline.svg?key_text=dev)](https://gitlab.com/aliby/aliby/-/commits/dev)\n[![coverage](https://gitlab.com/aliby/aliby/badges/dev/coverage.svg)](https://gitlab.com/aliby/aliby/-/commits/dev)\n\nEnd-to-end processing of cell microscopy time-lapses. ALIBY automates segmentation, tracking, lineage predictions, post-processing and report production. It leverages the existing Python ecosystem and open-source scientific software available to produce seamless and standardised pipelines.\n\n## Quickstart Documentation\nInstallation of [VS Studio](https://visualstudio.microsoft.com/downloads/#microsoft-visual-c-redistributable-for-visual-studio-2022) Native MacOS support for is under work, but you can use containers (e.g., Docker, Podman) in the meantime.\n\nTo analyse local data\n ```bash\npip install aliby \n ```\n Add any of the optional flags `omero` and `utils` (e.g., `pip install aliby[omero, utils]`). `omero` provides tools to connect with an OMERO server and `utils` provides visualisation, user interface and additional deep learning tools.\n  \nSee our [installation instructions]( https://aliby.readthedocs.io/en/latest/INSTALL.html ) for more details.\n\n### CLI\n\nIf installed via poetry, you have access to a Command Line Interface (CLI)\n\n ```bash\naliby-run --expt_id EXPT_PATH --distributed 4 --tps None\n ```\n\nAnd to run Omero servers, the basic arguments are shown:\n ```bash\n aliby-run --expt_id XXX --host SERVER.ADDRESS --user USER --password PASSWORD \n ```\n\nThe output is a folder with the original logfiles and a set of hdf5 files, one with the results of each multidimensional inside.\n\nFor more information, including available options, see the page on [running the analysis pipeline](https://aliby.readthedocs.io/en/latest/PIPELINE.html)\n\n## Using specific components\n\n### Access raw data\n\nALIBY\'s tooling can also be used as an interface to OMERO servers, for example, to fetch a brightfield channel.\n ```python\nfrom aliby.io.omero import Dataset, Image\n\nserver_info= {\n            "host": "host_address",\n            "username": "user",\n            "password": "xxxxxx"}\nexpt_id = XXXX\ntps = [0, 1] # Subset of positions to get.\n\nwith Dataset(expt_id, **server_info) as conn:\n    image_ids = conn.get_images()\n\n#To get the first position\nwith Image(list(image_ids.values())[0], **server_info) as image:\n    dimg = image.data\n    imgs = dimg[tps, image.metadata["channels"].index("Brightfield"), 2, ...].compute()\n    # tps timepoints, Brightfield channel, z=2, all x,y\n```\n\n### Tiling the raw data\n\nA `Tiler` object performs trap registration. It may be built in different ways but the simplest one is using an image and a the default parameters set.\n\n```python\nfrom aliby.tile.tiler import Tiler, TilerParameters\nwith Image(list(image_ids.values())[0], **server_info) as image:\n    tiler = Tiler.from_image(image, TilerParameters.default())\n    tiler.run_tp(0)\n```\n\nThe initialisation should take a few seconds, as it needs to align the images\nin time.\n\nIt fetches the metadata from the Image object, and uses the TilerParameters values (all Processes in aliby depend on an associated Parameters class, which is in essence a dictionary turned into a class.)\n\n#### Get a timelapse for a given tile (remote connection)\n```python\nfpath = "h5/location"\n\ntile_id = 9\ntrange = range(0, 10)\nncols = 8\n\nriv = remoteImageViewer(fpath)\ntrap_tps = [riv.tiler.get_tiles_timepoint(tile_id, t) for t in trange] \n\n# You can also access labelled traps\nm_ts = riv.get_labelled_trap(tile_id=0, tps=[0])\n\n# And plot them directly\nriv.plot_labelled_trap(trap_id=0, channels=[0, 1, 2, 3], trange=range(10))\n```\n\nDepending on the network speed can take several seconds at the moment.\nFor a speed-up: take fewer z-positions if you can.\n\n#### Get the tiles for a given time point\nAlternatively, if you want to get all the traps at a given timepoint:\n\n```python\ntimepoint = (4,6)\ntiler.get_tiles_timepoint(timepoint, channels=None,\n                                z=[0,1,2,3,4])\n```\n\n\n### Contributing\nSee [CONTRIBUTING](https://aliby.readthedocs.io/en/latest/INSTALL.html) on how to help out or get involved.\n',
    'author': 'Alan Munoz',
    'author_email': 'alan.munoz@ed.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import typing as t
from itertools import takewhile

import numpy as np
import pandas as pd
from tqdm import tqdm

from agora.abc import ParametersABC, ProcessABC
from agora.io.cells import Cells
from agora.io.signal import Signal
from agora.io.writer import Writer
from agora.utils.indexing import (
    _3d_index_to_2d,
    _assoc_indices_to_3d,
)
from agora.utils.merge import merge_association
from postprocessor.core.abc import get_parameters, get_process
from postprocessor.core.lineageprocess import (
    LineageProcess,
    LineageProcessParameters,
)
from postprocessor.core.reshapers.merger import Merger, MergerParameters
from postprocessor.core.reshapers.picker import Picker, PickerParameters


class PostProcessorParameters(ParametersABC):
    """
    Anthology of parameters used for postprocessing
    :merger:
    :picker: parameters for picker
    :processes: list processes:[objectives], 'processes' are defined in ./processes/
        while objectives are relative or absolute paths to datasets. If relative paths the
        post-processed addresses are used. The order of processes matters.

    Supply parameters for picker, merger, and processes.
    The order of processes matters.

    'processes' are defined in ./processes/ while objectives are relative or absolute paths to datasets. If relative paths the post-processed addresses are used.
    """

    def __init__(
        self,
        targets: t.Dict = {},
        param_sets: t.Dict = {},
        outpaths: t.Dict = {},
    ):
        self.targets = targets
        self.param_sets = param_sets
        self.outpaths = outpaths

    def __getitem__(self, item):
        return getattr(self, item)

    @classmethod
    def default(cls, kind=[]):
        """
        Include buddings and bud_metric and estimates of their time derivatives.

        Parameters
        ----------
        kind: list of str
            If "ph_batman" included, add targets for experiments using pHlourin.
        """
        # each subitem specifies the function to be called and the location
        # on the h5 file to be written
        targets = {
            "prepost": {
                "merger": "/extraction/general/None/area",
                "picker": ["/extraction/general/None/area"],
            },
            "processes": [
                [
                    "buddings",
                    ["/extraction/general/None/volume"],
                ],
                [
                    "dsignal",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "bud_metric",
                    [
                        "/extraction/general/None/volume",
                    ],
                ],
                [
                    "dsignal",
                    [
                        "/postprocessing/bud_metric/extraction_general_None_volume",
                    ],
                ],
            ],
        }
        param_sets = {
            "prepost": {
                "merger": MergerParameters.default(),
                "picker": PickerParameters.default(),
            }
        }
        outpaths = {}
        outpaths["aggregate"] = "/postprocessing/experiment_wide/aggregated/"
        # pHlourin experiments are special
        if "ph_batman" in kind:
            targets["processes"]["dsignal"].append(
                [
                    "/extraction/em_ratio/np_max/mean",
                    "/extraction/em_ratio/np_max/median",
                    "/extraction/em_ratio_bgsub/np_max/mean",
                    "/extraction/em_ratio_bgsub/np_max/median",
                ]
            )
            targets["processes"]["aggregate"].append(
                [
                    [
                        "/extraction/em_ratio/np_max/mean",
                        "/extraction/em_ratio/np_max/median",
                        "/extraction/em_ratio_bgsub/np_max/mean",
                        "/extraction/em_ratio_bgsub/np_max/median",
                        "/extraction/gsum/np_max/median",
                        "/extraction/gsum/np_max/mean",
                    ]
                ],
            )
        return cls(targets=targets, param_sets=param_sets, outpaths=outpaths)


class PostProcessor(ProcessABC):
    def __init__(self, filename, parameters):
        """
        Initialise PostProcessor

        Parameters
        ----------
        filename: str or PosixPath
            Name of h5 file.
        parameters: PostProcessorParameters object
            An instance of PostProcessorParameters.
        """
        super().__init__(parameters)
        self._filename = filename
        self._signal = Signal(filename)
        self._writer = Writer(filename)
        # parameters for merger and picker
        dicted_params = {
            i: parameters["param_sets"]["prepost"][i]
            for i in ["merger", "picker"]
        }
        for k in dicted_params.keys():
            if not isinstance(dicted_params[k], dict):
                dicted_params[k] = dicted_params[k].to_dict()
        # merger and picker
        self.merger = Merger(
            MergerParameters.from_dict(dicted_params["merger"])
        )
        self.picker = Picker(
            PickerParameters.from_dict(dicted_params["picker"]),
            cells=Cells.from_source(filename),
        )
        # processes, such as buddings
        self.classfun = {
            process: get_process(process)
            for process, _ in parameters["targets"]["processes"]
        }
        # parameters for the process in classfun
        self.parameters_classfun = {
            process: get_parameters(process)
            for process, _ in parameters["targets"]["processes"]
        }
        # locations to be written in the h5 file
        self.targets = parameters["targets"]

    def run_prepost(self):
        """Using picker, get and write lineages, returning mothers and daughters."""
        """Important processes run before normal post-processing ones"""
        record = self._signal.get_raw(self.targets["prepost"]["merger"])
        merges = np.array(self.merger.run(record), dtype=int)

        self._writer.write(
            "modifiers/merges", data=[np.array(x) for x in merges]
        )

        lineage = _assoc_indices_to_3d(self.picker.cells.mothers_daughters)
        lineage_merged = []

        if merges.any():  # Update lineages after merge events

            merged_indices = merge_association(lineage, merges)
            # Remove repeated labels post-merging
            lineage_merged = np.unique(merged_indices, axis=0)

        self.lineage = _3d_index_to_2d(
            lineage_merged if len(lineage_merged) else lineage
        )
        self._writer.write(
            "modifiers/lineage_merged", _3d_index_to_2d(lineage_merged)
        )

        picked_indices = self.picker.run(
            self._signal[self.targets["prepost"]["picker"][0]]
        )
        if picked_indices.any():
            self._writer.write(
                "modifiers/picks",
                data=pd.MultiIndex.from_arrays(
                    picked_indices.T,
                    # names=["trap", "cell_label", "mother_label"],
                    names=["trap", "cell_label"],
                ),
                overwrite="overwrite",
            )

    @staticmethod
    def pick_mother(a, b):
        """Update the mother id following this priorities:

        The mother has a lower id
        """
        x = max(a, b)
        if min([a, b]):
            x = [a, b][np.argmin([a, b])]
        return x

    def run(self):
        """
        Write the results to the h5 file.
        Processes include identifying buddings and finding bud metrics.
        """
        # run merger, picker, and find lineages
        self.run_prepost()
        # run processes
        for process, datasets in tqdm(self.targets["processes"]):
            if process in self.parameters["param_sets"].get("processes", {}):
                # parameters already assigned
                parameters = self.parameters_classfun[process](
                    self.parameters[process]
                )
            else:
                # assign parameters
                parameters = self.parameters_classfun[process].default()
            # load process
            loaded_process = self.classfun[process](parameters)
            if isinstance(parameters, LineageProcessParameters):
                loaded_process.lineage = self.lineage

            # apply process to each dataset
            for dataset in datasets:
                self.run_process(dataset, process, loaded_process)

    def run_process(self, dataset, process, loaded_process):
        """Run process on a single dataset and write the result."""
        # define signal
        if isinstance(dataset, list):
            # multisignal process
            signal = [self._signal[d] for d in dataset]
        elif isinstance(dataset, str):
            signal = self._signal[dataset]
        else:
            raise ("Incorrect dataset")
        # run process on signal
        if len(signal) and (
            not isinstance(loaded_process, LineageProcess)
            or len(loaded_process.lineage)
        ):
            result = loaded_process.run(signal)
        else:
            result = pd.DataFrame(
                [], columns=signal.columns, index=signal.index
            )
            result.columns.names = ["timepoint"]
        # define outpath, where result will be written
        if process in self.parameters["outpaths"]:
            outpath = self.parameters["outpaths"][process]
        elif isinstance(dataset, list):
            # no outpath is defined
            # place the result in the minimum common branch of all signals
            prefix = "".join(
                c[0]
                for c in takewhile(
                    lambda x: all(x[0] == y for y in x), zip(*dataset)
                )
            )
            outpath = (
                prefix
                + "_".join(  # TODO check that it always finishes in '/'
                    [d[len(prefix) :].replace("/", "_") for d in dataset]
                )
            )
        elif isinstance(dataset, str):
            outpath = dataset[1:].replace("/", "_")
        else:
            raise ("Outpath not defined", type(dataset))
        # add postprocessing to outpath when required
        if process not in self.parameters["outpaths"]:
            outpath = "/postprocessing/" + process + "/" + outpath
        # write result
        if isinstance(result, dict):
            # multiple Signals as output
            for k, v in result.items():
                self.write_result(
                    outpath + f"/{k}",
                    v,
                    metadata={},
                )
        else:
            # a single Signal as output
            self.write_result(
                outpath,
                result,
                metadata={},
            )

    def write_result(
        self,
        path: str,
        result: t.Union[t.List, pd.DataFrame, np.ndarray],
        metadata: t.Dict,
    ):
        self._writer.write(path, result, meta=metadata, overwrite="overwrite")



"""
Script in development
"""

# /usr/bin/env python3
import re
import warnings
from abc import abstractmethod
from collections import Counter
from pathlib import Path
from typing import Dict, Iterable, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from agora.abc import ProcessABC
from matplotlib.backends.backend_pdf import PdfPages
from numpy import ndarray
from scipy.signal import find_peaks

from postprocessor.grouper import NameGrouper

sns.set_style("darkgrid")


# Main dataframe structure

# | position | group | ntraps |robustness index | initial_ncells | final_ncells
# dir = "/home/alan/Documents/dev/skeletons/data/2021_06_15_pypipeline_unit_test_00/2021_06_15_pypipeline_unit_test_00/"
# dir = "/home/alan/Documents/dev/libs/aliby/data/2021_08_24_2Raf_00/2021_08_24_2Raf_00/"
# dirs = [
#     "16543_2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_01",
#     "16545_2019_07_16_aggregates_CTP_switch_2_0glu_0_0glu_URA7young_URA8young_URA8old_secondRun_01",
#     "18069_2019_12_05_aggregates_updownshift_2_0_2_URA8_URA7H360A_URA7H360R_00",
#     "18616_2020_02_20_protAgg_downUpShift_2_0_2_Ura8_Ura8HA_Ura8HR_01",
#     "18617_2020_02_21_protAgg_downUpShift_2_0_2_pHluorin_Ura7HA_Ura7HR_00",
#     "19129_2020_09_06_DownUpshift_2_0_2_glu_ura_mig1msn2_phluorin_00",
#     "19144_2020_09_07_DownUpshift_2_0_2_glu_ura_mig1msn2_phluorin_secondRound_00",
#     "19169_2020_09_09_downUpshift_2_0_2_glu_ura8_phl_mig1_phl_msn2_03",
#     "19199_2020_09_29_downUpshift_2_0_2_glu_ura8_ura8h360a_ura8h360r_00",
#     "19203_2020_09_30_downUpshift_twice_2_0_2_glu_ura8_ura8h360a_ura8h360r_00",
#     "19207_2020_10_01_exp_00",
#     "19232_2020_10_02_downUpshift_twice_2_0_2_glu_ura8_phluorinMsn2_phluorinMig1_01",
#     "19307_2020_10_22_downUpshift_2_01_2_glucose_dual_pH__dot6_nrg1_tod6__00",
#     "19310_2020_10_22_downUpshift_2_0_2_glu_dual_phluorin__glt1_psa1_ura7__thrice_00",
#     "19311_2020_10_23_downUpshift_2_0_2_glu_dual_phluorin__glt1_psa1_ura7__twice__04",
#     "19328_2020_10_31_downUpshift_four_2_0_2_glu_dual_phl__glt1_ura8_ura8__00",
#     "19329_2020_11_01_exp_00",
#     "19333_2020_11_02_downUpshift_2_0_2_glu_ura7_ura7ha_ura7hr_00",
#     "19334_2020_11_02_downUpshift_2_0_2_glu_ura8_ura8ha_ura8hr_00",
#     "19447_2020_11_18_downUpshift_2_0_2_glu_gcd2_gcd6_gcd7__02",
#     "19810_2021_02_21_ToxicityTest_00",
#     "19993_2021_06_15_pypipeline_unit_test_00",
#     "19996_2021_06_27_ph_calibration_dual_phl_ura8_5_04_5_83_7_69_7_13_6_59__01",
#     "20419_2021_11_02_dose_response_raf_05_075_2_glu_005_2_constantMedia_00",
# ]
# outdir = "/home/alan/Documents/dev/skeletons/data"
# dirs = Path(outdir).glob("*ph*")


# from abc import abstractclassmethod, abstractmethod


# group_pos_trap_ncells = (
#     concat.dropna().groupby(["group", "position", "trap"]).apply(len)
# )
# group_pos_trapswcell = (
#     group_pos_trap_ncells.dropna().groupby(["group", "position"]).apply(len)
# )


class Meta:
    """Convenience class to fetch data from hdf5 file."""

    def __init__(self, filename):
        self.filename = filename

    @property
    def ntimepoints(self):
        with h5py.File(self.filename, "r") as f:
            return f.attrs["time_settings/ntimepoints"][0]


class Compiler(ProcessABC):
    # def __init__(self, parameters):
    #     super().__init__(parameters)

    @abstractmethod
    def load_data(self):
        """Abstract function that must be reimplemented."""
        pass

    @abstractmethod
    def run():
        pass


class ExperimentCompiler(Compiler):
    def __init__(self, CompilerParameters, exp_path: Path):
        super().__init__(CompilerParameters)
        self.load_data(exp_path)

    def run(self):
        return {
            method: getattr(self, "compile_" + method)()
            for method in (
                "slice",
                "slices",
                "delta_traps",
                "pertrap_metric",
                "ncells",
                "last_valid_tp",
                "stages_dmetric",
                "fluorescence",
            )
        }

    def load_data(self, path: Path):
        self.grouper = NameGrouper(path)
        self.meta = Meta(self.grouper.files[0])

    @property
    def ntraps(self) -> dict:
        """Get the number of traps in each position.

        Returns ------- dict str -> int  Examples -------- FIXME: Add
        docs.
        """
        return {
            pos: coords.shape[0]
            for pos, coords in self.grouper.traplocs().items()
        }

    def concat_signal(self, sigloc=None, mode=None, **kwargs) -> pd.DataFrame:
        if sigloc is None:
            sigloc = "extraction/general/None/volume"
        self.sigloc = sigloc

        if mode is None:
            mode = "retained"

        if not hasattr(self, "_concat") or self.sigloc != sigloc:
            self._concat = self.grouper.concat_signal(
                self.sigloc, mode=mode, **kwargs
            )

        return self._concat

    def get_tp(self, sigloc=None, tp=None, mode=None, **kwargs) -> pd.Series:
        if tp is None:
            tp = 0

        if mode is None:
            mode = True

        return self.concat_signal(sigloc=sigloc, mode=mode, **kwargs).iloc[
            :, tp
        ]

    def count_cells(
        self,
        signal="extraction/general/None/volume",
        mode="raw",
        **kwargs,
    ):
        df = self.grouper.concat_signal(signal, mode=mode, **kwargs)
        df = df.groupby(["group", "position", "trap"]).count()
        df[df == 0] = np.nan
        return df

    def compile_dmetrics(self, stages=None):
        """Generate dataframe with dVol metrics without major cell picking."""
        names_signals = {
            "dvol": "postprocessing/dsignal/postprocessing_savgol_extraction_general_None_volume",
            "bud_dvol": "postprocessing/bud_metric/postprocessing_dsignal_postprocessing_savgol_extraction_general_None_volume",
        }
        names_signals = {
            "dvol": "postprocessing/dsignal/postprocessing_savgol_extraction_general_None_volume",
            "bud_dvol": "postprocessing/bud_metric/postprocessing_dsignal_postprocessing_savgol_extraction_general_None_volume",
            "buddings": "postprocessing/buddings/extraction_general_None_volume",
        }
        operations = {
            "dvol": ("dvol", "max"),
            "bud_dvol": ("bud_dvol", "max"),
            "buddings": ("buddings", "sum"),
            "buddings_mean": ("buddings", "mean"),
        }

        input_signals = {
            k: self.grouper.concat_signal(v) for k, v in names_signals.items()
        }

        ids = input_signals["buddings"].index
        for v in input_signals.values():
            ids = ids.intersection(v.index)

        if stages:

            def process_dfs(dfs, rng):
                return pd.DataFrame(
                    {
                        k: getattr(dfs[sig].loc(axis=1)[rng].loc[ids], op)(
                            axis=1
                        )
                        if isinstance(op, str)
                        else dfs[sig].loc[ids].apply(op, axis=1)
                        for k, (sig, op) in operations.items()
                    }
                )

            # Note that all input_signals columns must be the same
            col_vals = list(input_signals.values())[0].columns
            stages_dfs = {"Full": process_dfs(input_signals, col_vals)}
            for k, rng in stages:
                stage_df = process_dfs(input_signals, col_vals[rng])
                stages_dfs[k] = stage_df

        concat = pd.concat([x.reset_index() for x in stages_dfs.values()])
        concat["stage"] = np.array(
            [
                np.repeat(x, len(concat) // len(stages_dfs))
                for x in stages_dfs.keys()
            ]
        ).flatten()

        return (
            concat.set_index(["group", "position", "trap", "cell_label"])
            .melt("stage", ignore_index=False, var_name="growth_metric")
            .reset_index()
        )

    def compile_stages_dmetric(self):
        stages = self.get_stages()
        return self.compile_dmetrics(stages=stages)

    def get_stages(self):
        """Use the metadata to give a prediction of the media being pumped at
        each time point. Works for traditional metadata (pre-fluigent).

        Returns: ------ A list of tuples where in each the first value
        is the active     pump's contents and the second its associated
        range of time points
        """
        fpath = list(self.grouper.signals.values())[0].filename
        with h5py.File(fpath, "r") as f:
            tinterval = f.attrs.get("time_settings/timeinterval", None)[0]
            tnorm = tinterval / 60
            switch_times = f.attrs.get("switchtimes", None) / tnorm
            last_tp = (
                f.attrs.get("time_settings/totaltime", None)[0] / tinterval
            )
            pump_contents = f.attrs.get("pumpinit/contents", None)
            init_frate = f.attrs.get("pumpinit/flowrate", None)
            prate = f.attrs.get("pumprate", None)
            main_pump = np.array((init_frate.argmax(), *prate.argmax(axis=0)))

            intervals = np.array((0, *switch_times, last_tp), dtype=int)

            extracted_tps = self.grouper.ntimepoints
            stages = [  # Only add intervals with length larger than zero
                (
                    ": ".join((str(i + 1), pump_contents[p_id])),
                    range(intervals[i], min(intervals[i + 1], extracted_tps)),
                )
                for i, p_id in enumerate(main_pump)
                if (intervals[i + 1] > intervals[i])
            ]
            return stages

    def compile_growth_metrics(
        self,
        min_nbuddings: int = 2,
    ):
        """Filter mothers with n number of buddings and get their metrics.

        Select cells with at least two recorded buddings
        """
        names_signals = {
            "dvol": "postprocessing/dsignal/postprocessing_savgol_extraction_general_None_volume",
            "bud_dvol": "postprocessing/bud_metric/postprocessing_dsignal_postprocessing_savgol_extraction_general_None_volume",
            "buddings": "postprocessing/buddings/extraction_general_None_volume",
        }
        operations = {
            "dvol": ("dvol", "max"),
            "bud_dvol": ("bud_dvol", "max"),
            "buddings": ("buddings", "sum"),
            "cycle_length_mean": (
                "buddings",
                lambda x: bn.nanmean(np.diff(np.where(x)[0])),
            ),
            "cycle_length_min": (
                "buddings",
                lambda x: bn.nanmin(np.diff(np.where(x)[0])),
            ),
            "cycle_length_median": (
                "buddings",
                lambda x: np.nanmedian(np.diff(np.where(x)[0])),
            ),
        }

        input_signals = {
            k: self.grouper.concat_signal(v) for k, v in names_signals.items()
        }
        ids = self.get_shared_ids(input_signals, min_nbuddings=min_nbuddings)

        compiled_df = pd.DataFrame(
            {
                k: getattr(input_signals[sig].loc[ids], op)(axis=1)
                if isinstance(op, str)
                else input_signals[sig].loc[ids].apply(op, axis=1)
                for k, (sig, op) in operations.items()
            }
        )
        return compiled_df

    def get_shared_ids(
        self, input_signals: Dict[str, pd.DataFrame], min_nbuddings: int = None
    ):
        """Get the intersection id of multiple signals.

        "buddings" must be one the keys in input_signals to use the
        argument min_nbuddings.
        """
        ids = list(input_signals.values())[0].index
        if min_nbuddings is not None:
            ids = (
                input_signals["buddings"]
                .loc[input_signals["buddings"].sum(axis=1) >= min_nbuddings]
                .index
            )
        for v in input_signals.values():
            ids = ids.intersection(v.index)

        return ids

    def compile_ncells(self):
        df = self.count_cells()
        df = df.melt(ignore_index=False)
        df.columns = ["timepoint", "ncells_pertrap"]

        return df

    def compile_last_valid_tp(self) -> pd.Series:
        """Last valid timepoint per position."""
        df = self.count_cells()
        df = df.apply(lambda x: x.last_valid_index(), axis=1)
        df = df.groupby(["group", "position"]).max()

        return df

    def compile_slices(self, nslices=2, **kwargs):
        tps = [
            min(
                i * (self.grouper.ntimepoints // nslices),
                self.grouper.ntimepoints - 1,
            )
            for i in range(nslices + 1)
        ]
        slices = [self.compile_slice(tp=tp, **kwargs) for tp in tps]
        slices_df = pd.concat(slices)

        slices_df["timepoint"] = np.concatenate(
            [np.repeat(tp, len(slice_df)) for tp, slice_df in zip(tps, slices)]
        )

        return slices_df

    def compile_slice_end(self, **kwargs):
        return self.compile_slice(tp=-1, **kwargs)

    def guess_metrics(self, metrics: Dict[str, Tuple[str]] = None):
        """First approach at autoselecting certain signals for automated
        analysis."""

        if metrics is None:
            metrics = {
                "GFP": ("median", "max5"),
                "mCherry": ("median", "max5"),
                # "general": ("eccentricity",),
                "Flavin": ("median",),
                "postprocessing/savgol": ("volume",),
                "dsignal/postprocessing_savgol": ("volume",),
                "bud_metric.*dsignal.*savgol": ("volume",),
                "ph_ratio": ("median",),
            }

        sigs = self.grouper.siglist
        selection = {
            ".".join((ch, metric)): sig
            for sig in sigs
            for ch, metric_set in metrics.items()
            for metric in metric_set
            if re.search("(?!.*bgsub).*".join((ch, metric)) + "$", sig)
        }
        return selection

    def compile_fluorescence(
        self,
        metrics: Dict[str, Tuple[str]] = None,
        norm: tuple = None,
        **kwargs,
    ):
        """Get a single signal per."""
        if norm is None:
            norm = (
                "GFP",
                "GFPFast",
                "ph_ratio",
                "Flavin",
                "Citrine",
                "mCherry",
            )

        selection = self.guess_metrics(metrics)

        input_signals = {
            k: self.grouper.concat_signal(v, **kwargs)
            for k, v in selection.items()
        }

        # ids = self.get_shared_ids(input_signals)

        to_concat = []

        def format_df(df):
            return df.melt(
                ignore_index=False, var_name="timepoint"
            ).reset_index()

        for k, v in input_signals.items():
            tmp_formatted = format_df(v)
            tmp_formatted["signal"] = k
            to_concat.append(tmp_formatted)
            if norm and k.split(".")[0] in norm:
                norm_v = v.subtract(v.min(axis=1), axis=0).div(
                    v.max(axis=1) - v.min(axis=1), axis=0
                )
                # norm_v = v.groupby(["position", "trap", "cell_label"]).transform(
                #     # lambda x: x - x.min() / (x.max() - x.min())
                #     lambda x: (x - x.min())
                #     / (x.max() - x.min())
                # )
                formatted = format_df(norm_v)
                formatted["signal"] = "norm_" + k
                to_concat.append(formatted)

        concated = pd.concat(to_concat, axis=0)

        return concated

    def compile_slice(
        self, sigloc=None, tp=None, metrics=None, mode=None, **kwargs
    ) -> pd.DataFrame:
        if sigloc is None:
            self.sigloc = "extraction/general/None/volume"

        if tp is None:
            tp = 0

        if metrics is None:
            metrics = ("max", "mean", "median", "count", "std", "sem")

        if mode is None:
            mode = True

        df = pd.concat(
            [
                getattr(
                    self.get_tp(sigloc=sigloc, tp=tp, mode=mode, **kwargs)
                    .groupby(["group", "position", "trap"])
                    .max()
                    .groupby(["group", "position"]),
                    met,
                )()
                for met in metrics
            ],
            axis=1,
        )

        df.columns = metrics

        merged = self.add_column(df, self.ntraps, name="ntraps")

        return merged

    @staticmethod
    def add_column(df: pd.DataFrame, new_values_d: dict, name="new_col"):
        if name in df.columns:
            warnings.warn(
                "ExpCompiler: Replacing existing column in compilation"
            )
        df[name] = [
            new_values_d[pos] for pos in df.index.get_level_values("position")
        ]

        return df

    @staticmethod
    def traploc_diffs(traplocs: ndarray) -> list:
        """Obtain metrics for trap localisation.

        Parameters ---------- traplocs : ndarray     (x,2) 2-dimensional
        array with the x,y coordinates of traps in each     column
        Examples -------- FIXME: Add docs.
        """
        signal = np.zeros((traplocs.max(), 2))
        for i in range(2):
            counts = Counter(traplocs[:, i])
            for j, v in counts.items():
                signal[j - 1, i] = v

        diffs = [
            np.diff(x)
            for x in np.apply_along_axis(find_peaks, 0, signal, distance=10)[0]
        ]
        return diffs

    def compile_delta_traps(self):
        group_names = self.grouper.group_names
        tups = [
            (group_names[pos], pos, axis, val)
            for pos, coords in self.grouper.traplocs().items()
            for axis, vals in zip(("x", "y"), self.traploc_diffs(coords))
            for val in vals
        ]

        return pd.DataFrame(
            tups, columns=["group", "position", "axis", "value"]
        )

    def compile_pertrap_metric(
        self,
        ranges: Iterable[Iterable[int]] = [
            [0, -1],
        ],
        metric: str = "count",
    ):
        """Get the number of cells per trap present during the given ranges."""
        sig = self.concat_signal()

        for i, rngs in enumerate(ranges):
            for j, edge in enumerate(rngs):
                if edge < 0:
                    ranges[i][j] = sig.shape[1] - i + 1
        df = pd.concat(
            [
                self.get_filled_trapcounts(
                    sig.loc(axis=1)[slice(*rng)], metric=metric
                )
                for rng in ranges
            ],
            axis=1,
        )
        return df.astype(str)

    def get_filled_trapcounts(
        self, signal: pd.DataFrame, metric: str
    ) -> pd.Series:
        present = signal.apply(
            lambda x: (not x.first_valid_index())
            & (x.last_valid_index() == len(x) - 1),
            axis=1,
        )
        results = getattr(
            signal.loc[present]
            .iloc[:, 0]
            .groupby(["group", "position", "trap"]),
            metric,
        )()
        filled = self.fill_trapcount(results)
        return filled

    def fill_trapcount(
        self, srs: pd.Series, fill_value: Union[int, float] = 0
    ) -> pd.Series:
        """Fill the last level of a MultiIndex in a pd.Series.

        Use self to get the max number of traps per position and use
        this information to add rows with empty values (with plottings
        of distributions in mind)  Parameters ---------- srs : pd.Series
        Series with a pd.MultiIndex index self : ExperimentSelf
        class with 'ntraps' information that returns a dictionary with
        position     -> ntraps. fill_value : Union[int, float]     Value
        used to fill new rows.  Returns ------- pd.Series     Series
        with no numbers skipped on the last level.  Examples --------
        FIXME: Add docs.
        """

        all_sets = set(
            [
                (pos, i)
                for pos, ntraps in self.ntraps.items()
                for i in range(ntraps)
            ]
        )
        dif = all_sets.difference(
            set(
                zip(
                    *[
                        srs.index.get_level_values(i)
                        for i in ("position", "trap")
                    ]
                )
            ).difference()
        )
        new_indices = pd.MultiIndex.from_tuples(
            [
                (self.grouper.group_names[idx[0]], idx[0], np.uint(idx[1]))
                for idx in dif
            ]
        )
        new_indices = new_indices.set_levels(
            new_indices.levels[-1].astype(np.uint), level=-1
        )
        empty = pd.Series(fill_value, index=new_indices, name="ncells")
        return pd.concat((srs, empty))


class Reporter(object):
    """Manages Multiple pages to generate a report."""

    def __init__(
        self,
        data: Dict[str, pd.DataFrame],
        pages: dict = None,
        path: str = None,
    ):
        self.data = data

        if pages is None:
            pages = {
                "qa": self.gen_page_qa(),
                "growth": self.gen_page_growth(),
                "fluorescence": self.gen_page_fluorescence(),
            }
        self.pages = pages

        if path is not None:
            self.path = path

        self.porgs = {k: PageOrganiser(data, v) for k, v in pages.items()}

    @property
    def pdf(self):
        return self._pdf

    @pdf.setter
    def pdf(self, path: str):
        self._pdf = PdfPages(path)

    def plot_report(self, path: str = None):
        if path is None:
            path = self.path

        with PdfPages(path) as pdf:
            for page_org in list(self.porgs.values())[::-1]:
                page_org.plot_page()
                pdf.savefig(page_org.fig)
                # pdf.savefig()
                plt.close()

    @staticmethod
    def gen_page_qa():
        page_qc = (
            {
                "data": "slice",
                "func": "barplot",
                "args": ("ntraps", "position"),
                "kwargs": {"hue": "group", "palette": "muted"},
                "loc": (0, 0),
            },
            {
                "data": "delta_traps",
                "func": "barplot",
                "args": ("axis", "value"),
                "kwargs": {
                    "hue": "group",
                },
                "loc": (0, 1),
            },
            {
                "data": "slices",
                "func": "violinplot",
                "args": ("group", "median"),
                "kwargs": {
                    "hue": "timepoint",
                },
                "loc": (2, 1),
            },
            {
                "data": "pertrap_metric",
                "func": "histplot",
                "args": (0, None),
                "kwargs": {
                    "hue": "group",
                    "multiple": "dodge",
                    "discrete": True,
                },
                "loc": (2, 0),
            },
            {
                "data": "ncells",
                "func": "lineplot",
                "args": ("timepoint", "ncells_pertrap"),
                "kwargs": {
                    "hue": "group",
                },
                "loc": (1, 1),
            },
            {
                "data": "last_valid_tp",
                "func": "stripplot",
                "args": (0, "position"),
                "kwargs": {
                    "hue": "group",
                },
                "loc": (1, 0),
            },
        )
        return page_qc

    @staticmethod
    def gen_page_fluorescence():
        return (
            {
                "data": "fluorescence",
                "func": "relplot",
                "args": ("timepoint", "value"),
                "kwargs": {
                    "col": "signal",
                    "col_wrap": 2,
                    "hue": "group",
                    "facet_kws": {"sharey": False, "sharex": True},
                    "kind": "line",
                },
            },
        )

    def gen_page_cell_cell_corr():
        pass

    @staticmethod
    def gen_page_growth():
        return (
            {
                "data": "stages_dmetric",
                "func": "catplot",
                "args": ("stage", "value"),
                "kwargs": {
                    "hue": "group",
                    "col": "growth_metric",
                    "col_wrap": 2,
                    "kind": "box",
                    "sharey": False,
                },
            },
        )

    def gen_all_instructions(self):
        qa = self.gen_page_qa()
        growth = self.gen_page_growth()

        return (qa, growth)


class PageOrganiser(object):
    """Add multiple plots to a single page, wither using seaborn multiplots or
    manual GridSpec."""

    def __init__(
        self,
        data: Dict[str, pd.DataFrame],
        instruction_set: Iterable = None,
        grid_spec: tuple = None,
        fig_kws: dict = None,
    ):
        self.instruction_set = instruction_set
        self.data = {k: df for k, df in data.items()}

        self.single_fig = True
        if len(instruction_set) > 1:
            self.single_fig = False

        if not self.single_fig:  # Select grid_spec with location info
            if grid_spec is None:
                locs = np.array(
                    [x.get("loc", (0, 0)) for x in instruction_set]
                )
                grid_spec = locs.max(axis=0) + 1

            if fig_kws is None:
                self.fig = plt.figure(dpi=300)
                self.fig.set_size_inches(8.27, 11.69, forward=True)
                plt.figtext(0.02, 0.99, "", fontsize="small")
            self.gs = plt.GridSpec(*grid_spec, wspace=0.3, hspace=0.3)

            self.axes = {}
            reset_index = (
                lambda df: df.reset_index().sort_values("position")
                if isinstance(df.index, pd.core.indexes.multi.MultiIndex)
                else df.sort_values("position")
            )
            self.data = {k: reset_index(df) for k, df in self.data.items()}

    def place_plot(self, func, xloc=None, yloc=None, **kwargs):
        if xloc is None:
            xloc = 0
        if yloc is None:
            yloc = 0

        if (
            self.single_fig
        ):  # If plotting using a figure method using seaborn cols/rows
            self.g = func(**kwargs)
            self.axes = {
                ax.title.get_text().split("=")[-1][1:]: ax
                for ax in self.g.axes.flat
            }
            self.fig = self.g.fig
        else:
            self.axes[(xloc, yloc)] = self.fig.add_subplot(self.gs[xloc, yloc])
            func(
                ax=self.axes[(xloc, yloc)],
                **kwargs,
            )

        # Eye candy
        if np.any(  # If there is a long label, rotate them all
            [
                len(lbl.get_text()) > 8
                for ax in self.axes.values()
                for lbl in ax.get_xticklabels()
            ]
        ) and hasattr(self, "g"):
            for axes in self.g.axes.flat:
                _ = axes.set_xticklabels(
                    axes.get_xticklabels(),
                    rotation=15,
                    horizontalalignment="right",
                )

    def plot_page(
        self, instructions: Iterable[Dict[str, Union[str, Iterable]]] = None
    ):
        if instructions is None:
            instructions = self.instruction_set
        if isinstance(instructions, dict):
            how = (instructions,)

        for how in instructions:
            self.place_plot(
                self.gen_sns_wrapper(how),
                *how.get("loc", (None, None)),
            )

    def gen_sns_wrapper(self, how):
        def sns_wrapper(ax=None):
            kwargs = how.get("kwargs", {})
            if ax:
                kwargs["ax"] = ax
            elif "height" not in kwargs:
                ncols = kwargs.get("col_wrap", 1)
                if "col" in kwargs:
                    nrows = np.ceil(
                        len(np.unique(self.data[how["data"]][kwargs["col"]]))
                        / ncols
                    )
                else:
                    nrows = len(
                        np.unique(self.data[how["data"]][kwargs["row"]])
                    )

                kwargs["height"] = 11.7
                # kwargs["aspect"] = 8.27 / (11.7 / kwargs["col_wrap"])
                kwargs["aspect"] = (8.27 / ncols) / (kwargs["height"] / nrows)
            return getattr(sns, how["func"])(
                data=self.data[how["data"]],
                x=how["args"][0],
                y=how["args"][1],
                **kwargs,
            )

        return sns_wrapper


# fpath = "/home/alan/Documents/dev/skeletons/scripts/aggregates_exploration/18616_2020_02_20_protAgg_downUpShift_2_0_2_Ura8_Ura8HA_Ura8HR_01"
# # compiler = ExperimentCompiler(None, base_dir / dir)
# compiler = ExperimentCompiler(None, fpath)
# dfs = compiler.run()
# rep = Reporter(data=dfs, path=Path(fpath) / "report.pdf")
# rep.plot_report("./report.pdf")
# base_dir = Path("/home/alan/Documents/dev/skeletons/scripts/data/")
# for dir in dirs:
#     try:
#         compiler = ExperimentCompiler(None, base_dir / dir)
#         dfs = compiler.run()
#         rep = Reporter(data=dfs, path=base_dir / (dir + "/report.pdf"))
#         from time import time

#         rep.plot_report(base_dir / (dir + "/report.pdf"))
#     except Exception as e:
#         print("LOG:ERROR:", e)
# with open("errors.log", "a") as f:
#     f.write(e)



#!/usr/bin/env jupyter
import typing as t
from copy import copy

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans

from agora.utils.indexing import validate_association

index_row = t.Tuple[str, str, int, int]


def add_index_levels(
    df: pd.DataFrame, additional_ids: t.Dict[str, pd.Series] = {}
) -> pd.DataFrame:
    new_df = copy(df)
    for k, srs in additional_ids.items():
        assert len(srs) == len(
            new_df
        ), f"Series and new_df must match; sizes {len(srs)} and {len(new_df)}"
        new_df[k] = srs
        new_df.set_index(k, inplace=True, append=True)
    return new_df


def drop_level(
    df: pd.DataFrame,
    name: str = "mother_label",
    as_list: bool = True,
) -> t.Union[t.List[index_row], pd.Index]:
    """
    Drop index level.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe whose multiindex we will drop
    name : str
        Mame of index level to drop
    as_list : bool
        Whether to return as a list instead of an index
    """
    short_index = df.index.droplevel(name)
    if as_list:
        short_index = short_index.to_list()
    return short_index


def intersection_matrix(
    index1: pd.MultiIndex, index2: pd.MultiIndex
) -> np.ndarray:
    """Use casting to obtain the boolean mask of the intersection of two multi-indices."""
    indices = [index1, index2]
    for i in range(2):
        if hasattr(indices[i], "to_list"):
            indices[i]: t.List = indices[i].to_list()
        indices[i]: np.ndarray = np.array(indices[i])
    return (indices[0][..., None] == indices[1].T).all(axis=1)


def get_mother_ilocs_from_daughters(df: pd.DataFrame) -> np.ndarray:
    """Fetch mother locations in the index of df for all daughters in df."""
    daughter_ids = df.index[df.index.get_level_values("mother_label") > 0]
    mother_ilocs = intersection_matrix(
        daughter_ids.droplevel("cell_label"),
        drop_level(df, "mother_label", as_list=False),
    ).any(axis=0)
    return mother_ilocs


def get_mothers_from_another_df(whole_df: pd.DataFrame, da_df: pd.DataFrame):
    daughter_ids = da_df.index[
        da_df.index.get_level_values("mother_label") > 0
    ]
    mother_ilocs = intersection_matrix(
        daughter_ids.droplevel("cell_label"),
        drop_level(whole_df, "mother_label", as_list=False),
    ).any(axis=0)
    return mother_ilocs


def bidirectional_retainment_filter(
    df: pd.DataFrame,
    mothers_thresh: float = 0.8,
    daughters_thresh: int = 7,
) -> pd.DataFrame:
    """
    Retrieve families where mothers are present for more than a fraction of the experiment, and daughters for longer than some number of time-points.

    Parameters
    ----------
    df: pd.DataFrame
        Data
    mothers_thresh: float
        Minimum fraction of experiment's total duration for which mothers must be present.
    daughters_thresh: int
        Minimum number of time points for which daughters must be observed
    """
    # daughters
    all_daughters = df.loc[df.index.get_level_values("mother_label") > 0]
    # keep daughters observed sufficiently often
    retained_daughters = all_daughters.loc[
        all_daughters.notna().sum(axis=1) > daughters_thresh
    ]
    # fetch mother using existing daughters
    mothers = df.loc[get_mothers_from_another_df(df, retained_daughters)]
    # keep mothers present for at least a fraction of the experiment's duration
    retained_mothers = mothers.loc[
        mothers.notna().sum(axis=1) > mothers.shape[1] * mothers_thresh
    ]
    # drop daughters with no valid mothers
    final_da_mask = intersection_matrix(
        drop_level(retained_daughters, "cell_label", as_list=False),
        drop_level(retained_mothers, "mother_label", as_list=False),
    )
    final_daughters = retained_daughters.loc[final_da_mask.any(axis=1)]
    # join mothers and daughters and sort index
    return pd.concat((final_daughters, retained_mothers), axis=0).sort_index()


def melt_reset(df: pd.DataFrame, additional_ids: t.Dict[str, pd.Series] = {}):
    new_df = add_index_levels(df, additional_ids)

    return new_df.melt(
        ignore_index=False, var_name="time (minutes)", value_name="signal"
    ).reset_index()


# Drop cells that if used would reduce info the most
def filt_cluster(
    kymograph: pd.DataFrame,
    n: int = 2,
):
    mask = ~kymograph.iloc[:, kymograph.shape[1] // 2 :].isna().any(axis=1)
    informative = kymograph.loc[mask]

    clusters = cluster_kymograph(informative, n)

    return informative, clusters


def cluster_kymograph(kymograph: pd.DataFrame, n: int = 2):
    import bottleneck as bn
    from sklearn.cluster import KMeans

    # Normalise according to mean value of signal
    X = (
        kymograph.divide(bn.nanmean(kymograph, axis=1), axis=0)
        .dropna(axis=1)
        .values
    )
    km = KMeans(n, random_state=42).fit(X)
    clusters = km.predict(X)
    return clusters


def split_df(df, slices):
    return [df.iloc(axis=1)[slc] for slc in slices]


def slices_from_spans(spans: t.Tuple[int], df: pd.DataFrame) -> t.List[slice]:
    cumsum = np.cumsum(spans)

    slices = [
        slice(start, min(end, df.columns.get_level_values("time")[-1]))
        for start, end in zip(cumsum[:-1], cumsum[1:])
    ]
    return slices


def drop_mother_label(index: pd.MultiIndex) -> np.ndarray:
    no_mother_label = index
    if "mother_label" in index.names:
        no_mother_label = index.droplevel("mother_label")
    return np.array(no_mother_label.tolist())


def get_index_as_np(signal: pd.DataFrame):
    # Get mother labels from multiindex dataframe
    return np.array(signal.index.to_list())


def standard_filtering(
    raw: pd.DataFrame,
    lin: np.ndarray,
    presence_high: float = 0.8,
    presence_low: int = 7,
):
    # Get all mothers
    _, valid_indices = validate_association(
        lin, np.array(raw.index.to_list()), match_column=0
    )
    in_lineage = raw.loc[valid_indices]

    # Filter mothers by presence
    present = in_lineage.loc[
        in_lineage.notna().sum(axis=1) > (in_lineage.shape[1] * presence_high)
    ]

    # Get indices
    indices = np.array(present.index.to_list())
    to_cast = np.stack((lin[:, :2], lin[:, [0, 2]]), axis=1)
    ndin = to_cast[..., None] == indices.T[None, ...]

    # use indices to fetch all daughters
    valid_association = ndin.all(axis=2)[:, 0].any(axis=-1)

    # Remove repeats
    mothers, daughters = np.split(to_cast[valid_association], 2, axis=1)
    mothers = mothers[:, 0]
    daughters = daughters[:, 0]
    d_m_dict = {tuple(d): m[-1] for m, d in zip(mothers, daughters)}

    # assuming unique sorts
    raw_mothers = raw.loc[_as_tuples(mothers)]
    raw_mothers["mother_label"] = 0
    raw_daughters = raw.loc[_as_tuples(daughters)]
    raw_daughters["mother_label"] = d_m_dict.values()
    concat = pd.concat((raw_mothers, raw_daughters)).sort_index()
    concat.set_index("mother_label", append=True, inplace=True)

    # Last filter to remove tracklets that are too short
    removed_buds = concat.notna().sum(axis=1) <= presence_low
    filt = concat.loc[~removed_buds]

    # We check that no mothers are left child-less
    m_d_dict = {tuple(m): [] for m in mothers}
    for (trap, d), m in d_m_dict.items():
        m_d_dict[(trap, m)].append(d)

    for trap, daughter, mother in concat.index[removed_buds]:
        idx_to_delete = m_d_dict[(trap, mother)].index(daughter)
        del m_d_dict[(trap, mother)][idx_to_delete]

    bud_free = []
    for m, d in m_d_dict.items():
        if not d:
            bud_free.append(m)

    final_result = filt.drop(bud_free)

    # In the end, we get the mothers present for more than {presence_lineage1}% of the experiment
    # and their tracklets present for more than {presence_lineage2} time-points
    return final_result



#!/usr/bin/env python
"""
TrackerCoordinator class to coordinate cell tracking and bud assignment.
"""
import pickle
import typing as t
from collections import Counter
from os.path import dirname, join
from pathlib import Path

import numpy as np
from scipy.optimize import linear_sum_assignment
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC

from agora.track_abc import FeatureCalculator

models_path = join(dirname(__file__), "../models")


class CellTracker(FeatureCalculator):
    """
    Class used to manage cell tracking. You can call it using an existing model or
    use the inherited CellTrainer to get a new one.

    Initialization parameters:

    :model: sklearn.ensemble.RandomForestClassifier object
    :trapfeats: Features to manually calculate within a trap
    :extrafeats: Additional features to calculate
    :model: Model to use, if provided ignores all other args but threshs
    :bak_model: Backup mode to use when prediction is unsure
    :nstepsback: int Number of timepoints to go back
    :thresh: float Cut-off value to assume a cell is not new
    :low_thresh: Lower thresh for model switching
    :high_thresh: Higher thresh for model switching.
        Probabilities between these two thresholds summon the
        backup model.

    :aweights: area weight for barycentre calculations


    # Feature order in array features
    1. basic features
    2. trap features (within a trap)
    3. extra features (process between imgs)

    """

    def __init__(
        self,
        feats2use=None,
        trapfeats=None,
        extrafeats=None,
        model=None,
        bak_model=None,
        thresh=None,
        low_thresh=None,
        high_thresh=None,
        nstepsback=None,
        aweights=None,
        red_fun=None,
        max_distance=None,
        **kwargs,
    ):

        if trapfeats is None:
            trapfeats = ()

        if extrafeats is None:
            extrafeats = ()

        if type(model) is str or type(model) is Path:
            with open(Path(model), "rb") as f:
                model = pickle.load(f)

        if type(bak_model) is str or type(bak_model) is Path:
            with open(Path(bak_model), "rb") as f:
                bak_model = pickle.load(f)

        if aweights is None:
            self.aweights = None

        if feats2use is None:  # Ignore this block when training
            if model is None:
                model = self.load_model(models_path, "ct_XGBC_20220703_11.pkl")
            if bak_model is None:
                bak_model = self.load_model(
                    models_path, "ct_XGBC_20220703_10.pkl"
                )
            self.model = model
            self.bak_model = bak_model

            main_feats = model.all_ifeats
            bak_feats = bak_model.all_ifeats
            feats2use, trapfeats, extrafeats = [
                tuple(sorted(set(main).union(bak)))
                for main, bak in zip(main_feats, bak_feats)
            ]

        # Training AND non-training part
        super().__init__(
            feats2use, trapfeats=trapfeats, extrafeats=extrafeats, **kwargs
        )

        self.extrafeats = tuple(extrafeats)

        self.all_ofeats = self.outfeats + trapfeats + extrafeats

        self.noutfeats = len(self.all_ofeats)

        if hasattr(self, "bak_model"):  # Back to non-training only

            self.mainof_ids = [
                self.all_ofeats.index(f) for f in self.model.all_ofeats
            ]

            self.bakof_ids = [
                self.all_ofeats.index(f) for f in self.bak_model.all_ofeats
            ]

        if nstepsback is None:
            nstepsback = 3
        self.nstepsback = nstepsback

        if thresh is None:
            thresh = 0.5
        self.thresh = thresh
        if low_thresh is None:
            low_thresh = 0.4
        if high_thresh is None:
            high_thresh = 0.6
        self.low_thresh, self.high_thresh = low_thresh, high_thresh

        if red_fun is None:
            red_fun = np.nanmax
        self.red_fun = red_fun

        if max_distance is None:
            self.max_distance = max_distance

    def get_feats2use(self):
        """
        Return feats to be used from a loaded random forest model model
        """
        nfeats = get_nfeats_from_model(self.model)
        nfeats_bak = get_nfeats_from_model(self.bak_model)
        # max_nfeats = max((nfeats, nfeats_bak))

        return (switch_case_nfeats(nfeats), switch_case_nfeats(nfeats_bak))

    def calc_feat_ndarray(self, prev_feats, new_feats):
        """
        Calculate feature ndarray using two ndarrays of features.
        ---

        input

        :prev_feats: ndarray (ncells, nfeats) of timepoint 1
        :new_feats: ndarray (ncells, nfeats) of timepoint 2

        returns

        :n3darray: ndarray (ncells_prev, ncells_new, nfeats) containing a
            cell-wise substraction of the features in the input ndarrays.
        """
        if not (new_feats.any() and prev_feats.any()):
            return np.array([])

        n3darray = np.empty((len(prev_feats), len(new_feats), self.ntfeats))

        # print('self: ', self, ' self.ntfeats: ', self.ntfeats, ' featsshape: ', new_feats.shape)
        for i in range(self.ntfeats):
            n3darray[..., i] = np.subtract.outer(
                prev_feats[:, i], new_feats[:, i]
            )

        n3darray = self.calc_dtfeats(n3darray)

        return n3darray

    def calc_dtfeats(self, n3darray):
        """
        Calculates features obtained between timepoints, such as distance
        for every pair of cells from t1 to t2.
        ---

        input

        :n3darray: ndarray (ncells_prev, ncells_new, ntfeats) containing a
            cell-wise substraction of the features in the input ndarrays.

        returns

        :newarray: 3d array taking the features specified in self.outfeats and self.trapfeats
        and adding dtfeats
        """
        newarray = np.empty(n3darray.shape[:-1] + (self.noutfeats,))
        newarray[..., : len(self.outfeats)] = n3darray[
            ..., : len(self.outfeats)
        ]
        newarray[
            ..., len(self.outfeats) : len(self.outfeats) + len(self.trapfeats)
        ] = n3darray[..., len(self.out_merged) :]
        for i, feat in enumerate(self.all_ofeats):
            if feat == "distance":
                newarray[..., i] = np.sqrt(
                    n3darray[..., self.xind] ** 2
                    + n3darray[..., self.yind] ** 2
                )

        return newarray

    def assign_lbls(
        self,
        prob_backtrace: np.ndarray,
        prev_lbls: t.List[t.List[int]],
        red_fun=None,
    ):
        """Assign labels using a prediction matrix of nxmxl where n is the number
        of cells in the previous image, m the number of steps back considered
        and l in the new image. It assigns the
        number zero if it doesn't find the cell.
        ---
        input

        :prob_backtrace: Probability n x m x l array obtained as an output of rforest
        :prev_labels: List of cell labels for previous timepoint to be compared.
        :red_fun: Function used to collapse the previous timepoints into one.
            If none provided it uses maximum and ignores np.nans.

        returns

        :new_lbls: ndarray of newly assigned labels obtained, new cells as
        zero.
        """
        if red_fun is None:
            red_fun = self.red_fun

        new_lbls = np.zeros(prob_backtrace.shape[2], dtype=int)
        pred_matrix = red_fun(prob_backtrace, axis=1)

        if pred_matrix.any():
            # assign available hits
            row_ids, col_ids = linear_sum_assignment(-pred_matrix)
            for i, j in zip(row_ids, col_ids):
                if pred_matrix[i, j] > self.thresh:
                    new_lbls[j] = prev_lbls[i]

        return new_lbls

    def predict_proba_from_ndarray(
        self,
        array_3d: np.ndarray,
        model: str = None,
        boolean: bool = False,
        max_distance: float = None,
    ):
        """

        input

        :array_3d: (ncells_tp1, ncells_tp2, out_feats) ndarray
        :model: str, {'model', 'bak_model'} can force a unique model instead of an ensemble
        :boolean: bool, if False returns probability, if True returns prediction
        :max_distance: float Maximum distance (in um) to be considered. If None it uses the instance's value,
        if zero it skips checking distances.

        requires
        :self.model:
        :self.mainof_ids: list of indices corresponding to the main model's features
        :self.bakof_ids: list of indices corresponding to the backup model's features

        returns

        (ncells_tp1, ncells_tp2) ndarray with probabilities or prediction
            of cell identities depending on "boolean" arg.

        """

        if array_3d.size == 0:
            return np.array([])

        if model is None:
            model2use = self.model
            bak_model2use = self.bak_model
            bakof_ids = self.bakof_ids
            mainof_ids = self.mainof_ids
        else:
            model2use = getattr(self, "model")
            bak_model2use = model2use
            bakof_ids = [
                self.all_ofeats.index(f) for f in model2use.all_ofeats
            ]
            mainof_ids = [
                self.all_ofeats.index(f) for f in model2use.all_ofeats
            ]

        fun2use = "predict" if boolean else "predict_proba"
        predict_fun = getattr(model2use, fun2use)
        bak_pred_fun = getattr(bak_model2use, fun2use)

        if max_distance is None:
            max_distance = self.max_distance

        orig_shape = array_3d.shape[:2]

        # Ignore cells that are too far away to possibly be the same
        cells_near = np.ones(orig_shape, dtype=bool)
        if max_distance and set(self.all_ofeats).intersection(
            ("distance", "centroid-0", "centroid-1")
        ):
            if "distance" in self.all_ofeats:
                cells_near = (
                    array_3d[..., self.all_ofeats.index("distance")]
                    < max_distance
                )
            else:  # Calculate euclidean distance
                cells_near = (
                    np.sqrt(
                        array_3d[..., self.all_ofeats.index("centroid-0")]
                        + array_3d[..., self.all_ofeats.index("centroid-1")]
                    )
                    < max_distance
                )

        pred_matrix = np.zeros(orig_shape)
        prob = np.zeros(orig_shape)
        if cells_near.any():
            prob = predict_fun(array_3d[cells_near][:, mainof_ids])[:, 1]
            uncertain_dfeats = (self.low_thresh < prob) & (
                prob < self.high_thresh
            )
            if uncertain_dfeats.any():
                bak_prob = bak_pred_fun(
                    array_3d[cells_near][uncertain_dfeats][:, bakof_ids]
                )[:, 1]
                probs_compared = np.stack((prob[uncertain_dfeats], bak_prob))
                most_confident_proba = abs((probs_compared - 0.5)).argmax(
                    axis=0
                )
                prob[uncertain_dfeats] = probs_compared[
                    most_confident_proba, range(most_confident_proba.shape[0])
                ]

            pred_matrix[cells_near] = prob

        return pred_matrix

    def get_new_lbls(
        self,
        new_img,
        prev_lbls,
        prev_feats,
        max_lbl,
        new_feats=None,
        pixel_size=None,
        **kwargs,
    ):
        """
        Core function to calculate the new cell labels.

        ----

        input

        :new_img: ndarray (len, width, ncells) containing the cell outlines
        :max_lbl: int indicating the last assigned cell label
        :prev_feats: list of ndarrays of size (ncells x nfeatures)
        containing the features of previous timepoints
        :prev_lbls: list of list of ints corresponding to the cell labels in
            the previous timepoints
        :new_feats: (optional) Directly give a feature ndarray. It ignores
            new_img if given.
        :kwargs: Additional keyword values passed to self.predict_proba_from_ndarray

        returns

        :new_lbls: list of labels assigned to new timepoint
        :new_feats: list of ndarrays containing the updated features
        :new_max: updated max cell label assigned

        """

        if new_feats is None:
            new_feats = self.calc_feats_from_mask(new_img)

        if new_feats.any():
            if np.any([len(prev_feat) for prev_feat in prev_feats]):
                counts = Counter(
                    [lbl for lbl_set in prev_lbls for lbl in lbl_set]
                )
                lbls_order = list(counts.keys())
                probs = np.full(
                    (len(lbls_order), self.nstepsback, len(new_feats)), np.nan
                )

                for i, (lblset, prev_feat) in enumerate(
                    zip(prev_lbls, prev_feats)
                ):
                    if len(prev_feat):
                        feats_3darray = self.calc_feat_ndarray(
                            prev_feat, new_feats
                        )

                        pred_matrix = self.predict_proba_from_ndarray(
                            feats_3darray,
                            **kwargs,
                        )

                        for j, lbl in enumerate(lblset):
                            probs[lbls_order.index(lbl), i, :] = pred_matrix[
                                j, :
                            ]

                new_lbls = self.assign_lbls(probs, lbls_order)
                new_cells_pos = new_lbls == 0
                new_max = max_lbl + sum(new_cells_pos)
                new_lbls[new_cells_pos] = [*range(max_lbl + 1, new_max + 1)]

                # ensure that label output is consistently a list
                new_lbls = new_lbls.tolist()

            else:
                new_lbls = [*range(max_lbl + 1, max_lbl + len(new_feats) + 1)]

                new_max = max_lbl + len(new_feats)

        else:
            return ([], [], max_lbl)
        return (new_lbls, new_feats, new_max)

    def probabilities_from_impair(
        self,
        image_t1: np.ndarray,
        image_t2: np.ndarray,
        kwargs_feat_calc: dict = {},
        **kwargs,
    ):
        """
        Convenience function to test tracking between two time-points

        :image_t1: np.ndarray containing mask of first time-point
        :image_t2: np.ndarray containing mask of second time-point
        :kwargs_feat_calc: are passed to self.calc_feats_from_mask calls
        :kwargs: are passed to self.predict_proba_from_ndarray
        """
        feats_t1 = self.calc_feats_from_mask(image_t1, **kwargs_feat_calc)
        feats_t2 = self.calc_feats_from_mask(image_t2, **kwargs_feat_calc)

        probability_matrix = np.array([])
        if feats_t1.any() and feats_t2.any():
            feats_3darray = self.calc_feat_ndarray(feats_t1, feats_t2)

            probability_matrix = self.predict_proba_from_ndarray(
                feats_3darray, **kwargs
            )

        return probability_matrix

    # Step CellTracker
    def run_tp(
        self,
        masks: np.ndarray,
        state: t.Dict[str, t.Union[int, list]] = None,
        **kwargs,
    ) -> t.Dict[str, t.Union[t.List[int], t.Dict[str, t.Union[int, list]]]]:
        """Assign labels to new masks using a state dictionary.

        Parameters
        ----------
        masks : np.ndarray
            Cell masks to label.
        state : t.Dict[str, t.Union[int, list]]
            Dictionary containing maximum cell label, and previous cell labels
            and features for those cells.
        kwargs : keyword arguments
            Keyword arguments passed to self.get_new_lbls

        Returns
        -------
        t.Dict[str, t.Union[t.List[int], t.Dict[str, t.Union[int, list]]]]
            New labels and new state dictionary.

        Examples
        --------
        FIXME: Add example beyond the trivial one.

        import numpy as np
        from baby.tracker.core import celltracker
        from tqdm import tqdm

        # overlapping outlines are of shape (t,z,x,y)
        masks = np.zeros((5, 3, 20, 20), dtype=bool)
        masks[0, 0, 2:6, 2:6] = true
        masks[1:, 0, 13:14, 13:14] = true
        masks[:, 1, 8:12, 8:12] = true
        masks[:, 2, 14:18, 14:18] = true

        # 13um pixel size
        ct = celltracker(pixel_size=0.185)
        labels = []

        state = none
        for masks_tp in tqdm(masks):
            new_labels, state = ct.run_tp(masks_tp, state=state)
            labels.append(new_labels)

        # should result in state['cell_lbls']
        # [[1, 2, 3], [4, 2, 3], [4, 2, 3], [4, 2, 3], [4, 2, 3]]
        """
        if state is None:
            state = {}

        max_lbl = state.get("max_lbl", 0)
        cell_lbls = state.get("cell_lbls", [])
        prev_feats = state.get("prev_feats", [])

        # Get features for cells at this time point
        feats = self.calc_feats_from_mask(masks)

        lastn_lbls = cell_lbls[-self.nstepsback :]
        lastn_feats = prev_feats[-self.nstepsback :]

        new_lbls, _, max_lbl = self.get_new_lbls(
            masks, lastn_lbls, lastn_feats, max_lbl, feats, **kwargs
        )

        state = {
            "max_lbl": max_lbl,
            "cell_lbls": cell_lbls + [new_lbls],
            "prev_feats": prev_feats + [feats],
        }

        return (new_lbls, state)


# Helper functions


def switch_case_nfeats(nfeats):
    """
    Convenience TEMPORAL function to determine whether to use distance/location
    as a feature for tracking or not (nfeats=5 for no distance, 7 for distance)
    input
    number of feats

    returns
    list of main and extra feats based on the number of feats
    """
    main_feats = {
        4: [
            ("area", "minor_axis_length", "major_axis_length", "bbox_area"),
            (),
            (),
        ],
        5: [
            (
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                "perimeter",
            ),
            (),
            (),
        ],
        6: [
            (
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                "perimeter",
            ),
            (),
            ("distance",),
        ],
        # Including centroid
        # 7 : [('centroid', 'area', 'minor_axis_length', 'major_axis_length',
        #       'bbox_area', 'perimeter'), () , ()],
        7: [
            (
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                "perimeter",
            ),
            ("baryangle", "barydist"),
            (),
        ],
        8: [  # Minus centroid
            (
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                # "eccentricity",
                "equivalent_diameter",
                # "solidity",
                # "extent",
                "orientation",
                # "perimeter",
            ),
            ("baryangle", "barydist"),
            (),
        ],
        10: [  # Minus distance
            (
                "centroid",
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                # "eccentricity",
                "equivalent_diameter",
                # "solidity",
                # "extent",
                "orientation",
                # "perimeter",
            ),
            ("baryangle", "barydist"),
            (),
        ],
        11: [  # Minus computationally-expensive features
            (
                "centroid",
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                # "eccentricity",
                "equivalent_diameter",
                # "solidity",
                # "extent",
                "orientation",
                # "perimeter",
            ),
            ("baryangle", "barydist"),
            ("distance",),
        ],
        15: [  # All features
            (
                "centroid",
                "area",
                "minor_axis_length",
                "major_axis_length",
                "bbox_area",
                "eccentricity",
                "equivalent_diameter",
                "solidity",
                "extent",
                "orientation",
                "perimeter",
            ),
            ("baryangle", "barydist"),
            ("distance",),
        ],
    }

    assert nfeats in main_feats.keys(), "invalid nfeats"
    return main_feats.get(nfeats, [])


def get_nfeats_from_model(model) -> int:
    if isinstance(model, SVC):
        nfeats = model.support_vectors_.shape[-1]
    elif isinstance(model, RandomForestClassifier):
        nfeats = model.n_features_

    return nfeats



#!/usr/bin/env jupyter
"""
Convenience decorators to extend commonly-used methods or functions.
"""
import typing as t
from functools import wraps


def _first_arg_str_to_df(
    fn: t.Callable,
):
    """Enable Signal-like classes to convert strings to data sets."""
    @wraps(fn)
    def format_input(*args, **kwargs):
        cls = args[0]
        data = args[1]
        if isinstance(data, str):
            # get data from h5 file
            data = cls.get_raw(data)
        # replace path in the undecorated function with data
        return fn(cls, data, *args[2:], **kwargs)
    return format_input



#!/usr/bin/env python3

import igraph as ig
import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from sklearn.metrics.pairwise import euclidean_distances

from postprocessor.core.abc import PostProcessABC


class knngraphParameters(ParametersABC):
    """
    Parameters for the 'knngraph' process.
    """

    _defaults = {
        "n_neighbors": 10,
    }


class knngraph(PostProcessABC):
    """
    Process to get geometric graph from data/features, based on k-nearest
    neighbours.

    Methods
    -------
    run(signal: pd.DataFrame)
        Get graph.
    """

    def __init__(self, parameters: knngraphParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        """Get graph.

        Parameters
        ----------
        signal : pd.DataFrame
            Feature matrix.
        """
        distance_matrix = euclidean_distances(signal)
        distance_matrix_pruned = graph_prune(
            distance_matrix, self.n_neighbours
        )
        graph = ig.Graph.Weighted_Adjacency(
            distance_matrix_pruned.tolist(), mode="undirected"
        )
        graph.vs["strain"] = signal.index.get_level_values("strain")
        return graph


def graph_prune(distanceMatrix, neighbours):
    """
    Prunes a complete graph (input distance matrix), keeping at least a
    specified number of neighbours for each node.

    Parameters:
    -----------
    distanceMatrix = 2D numpy array
    neighbours = integer

    Return: Dij_pruned, a 2D numpy array, represents distance matrix of pruned
            graph
    """
    Dij_temp = distanceMatrix
    Adj = np.zeros(distanceMatrix.shape)
    for ii in range(distanceMatrix.shape[0]):
        idx = np.argsort(Dij_temp[ii, :])
        Adj[ii, idx[1]] = 1
        Adj[idx[1], ii] = 1
        for jj in range(neighbours):
            Adj[ii, idx[jj]] = 1
            Adj[idx[jj], ii] = 1
    Dij_pruned = Dij_temp * Adj
    return Dij_pruned
